#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""RS232 function library for Julabo circulators.
Tested on model FP51-SL.
"""
__author__ = "Dennis van Gils"
__authoremail__ = "vangils.dennis@gmail.com"
__url__ = "https://github.com/Dennis-van-Gils/python-dvg-devices"
__date__ = "12-06-2024"
__version__ = "1.5.0"
# pylint: disable=broad-except, missing-function-docstring, multiple-statements

import time
from typing import Union, Tuple

import serial
import numpy as np

from dvg_debug_functions import print_fancy_traceback as pft
from dvg_devices.BaseDevice import SerialDevice


# The manual states that
# - OUT commands should have a time gap > 250 ms. These are 'send' operations.
# - IN commands should have a time gap > 10 ms. These are 'query' operations.
# This module will enforce these time gaps.
DELAY_COMMAND_IN = 0.02  # [s] 0.02 tested stable
DELAY_COMMAND_OUT = 0.3  # [s] 0.3 tested stable


class Julabo_circulator(SerialDevice):
    class State:
        """Container for the process and measurement variables."""

        # fmt: off
        version: str = ""    # Version of the Julabo firmware
                             # FP51-SL: "JULABO HIGHTECH FL HL/SL VERSION 4.0"
        temp_unit: str = ""  # Temperature unit used by the Julabo  ("C"; "F")
        status   : Union[float, str] = np.nan  # Status or error message of the Julabo
        has_error: Union[float, bool] = np.nan  # True when status is a negative number
        running  : Union[float, bool] = np.nan  # Is the circulator running?

        setpoint_preset = np.nan # Active setpoint preset in the Julabo (1; 2; 3)
        setpoint = np.nan    # Read-out temp. setpoint of active preset [C; F]
        setpoint_1 = np.nan  # Read-out temp. setpoint preset #1        [C; F]
        setpoint_2 = np.nan  # Read-out temp. setpoint preset #2        [C; F]
        setpoint_3 = np.nan  # Read-out temp. setpoint preset #3        [C; F]
        bath_temp = np.nan   # Current bath temperature                 [C; F]
        pt100_temp = np.nan  # Current external Pt100 temperature       [C; F]

        over_temp = np.nan   # High-temperature warning limit           [C; F]
        sub_temp = np.nan    # Low-temperature warning limit            [C; F]

        # The Julabo has an independent temperature safety circuit. When the
        # safety sensor reading `SafeSens` is above the screw-set excess
        # temperature protection `SafeTemp`, the circulator will switch off.
        safe_sens = np.nan   # Safety sensor temperature reading        [C; F]
        safe_temp = np.nan   # Screw-set excess temperature protection  [C; F]

        # Time keeping to slow down communication per manual specs
        t_prev_out = np.nan  # Timestamp of previous OUT command [s]
        t_prev_in = np.nan   # Timestamp of previous IN command [s]
        # fmt: on

    def __init__(self, name="Julabo", long_name="Julabo circulator"):
        super().__init__(name=name, long_name=long_name)

        # Default serial settings
        self.serial_settings = {
            "baudrate": 4800,
            "bytesize": serial.SEVENBITS,
            "parity": serial.PARITY_EVEN,
            "timeout": 0.5,
            "write_timeout": 0.5,
        }
        self.set_read_termination("\r\n")
        self.set_write_termination("\r")

        self.set_ID_validation_query(
            ID_validation_query=self.ID_validation_query,
            valid_ID_broad="JULABO",
            valid_ID_specific=None,
        )

        # Container for the process and measurement variables
        self.state = self.State()

    # --------------------------------------------------------------------------
    #   ID_validation_query
    # --------------------------------------------------------------------------

    def ID_validation_query(self) -> Tuple[str, Union[str, None]]:
        # Strange Julabo quirk: The first query always times out
        try:
            self.query("VERSION")
        except Exception:
            pass  # Ignore the first time-out

        _success, reply = self.query("VERSION")
        if isinstance(reply, str):
            broad_reply = reply[:6]  # Expected: "JULABO"
            reply_specific = reply[7:]
        else:
            broad_reply = ""
            reply_specific = None

        return broad_reply, reply_specific

    # --------------------------------------------------------------------------
    #   begin
    # --------------------------------------------------------------------------

    def begin(self) -> bool:
        """This function should run directly after having established a
        connection to a Julabo. It retrieves the first readings and settings
        from the Julabo.

        Returns: True if successful, False otherwise.
        """

        success = True

        success &= self.query_version()
        success &= self.query_temp_unit()
        success &= self.query_sub_temp()
        success &= self.query_over_temp()
        success &= self.query_safe_temp()

        success &= self.query_running()
        success &= self.query_setpoint_preset()
        success &= self.query_setpoint()
        success &= self.query_setpoint_1()
        success &= self.query_setpoint_2()
        success &= self.query_setpoint_3()

        success &= self.query_bath_temp()
        success &= self.query_pt100_temp()
        success &= self.query_safe_sens()
        success &= self.query_status()

        return success

    # --------------------------------------------------------------------------
    #   turn_on/off
    # --------------------------------------------------------------------------

    def turn_off(self) -> bool:
        """Turn the Julabo off.

        Returns: True if successful, False otherwise.
        """

        if self.write_("OUT_MODE_05 0"):
            self.state.running = False
            return True

        return False

    def turn_on(self) -> bool:
        """Turn the Julabo on.

        Returns: True if successful, False otherwise.
        """

        if self.write_("OUT_MODE_05 1"):
            self.state.running = True
            return True

        return False

    # --------------------------------------------------------------------------
    #   set_sub_temp
    # --------------------------------------------------------------------------

    def set_sub_temp(self, value: float) -> bool:
        """Set the low-temperature warning limit. Subsequently, the Julabo is
        queried for the obtained value, which might be different than the one
        requested.

        Returns: True if all communication was successful, False otherwise.
        """

        try:
            value = float(value)
        except (TypeError, ValueError) as err:
            pft(err)
            return False

        if self.write_(f"OUT_SP_04 {value:.2f}"):
            return self.query_sub_temp()

        return False

    # --------------------------------------------------------------------------
    #   set_over_temp
    # --------------------------------------------------------------------------

    def set_over_temp(self, value: float) -> bool:
        """Set the high-temperature warning limit. Subsequently, the Julabo is
        queried for the obtained value, which might be different than the one
        requested.

        Returns: True if all communication was successful, False otherwise.
        """

        try:
            value = float(value)
        except (TypeError, ValueError) as err:
            pft(err)
            return False

        if self.write_(f"OUT_SP_03 {value:.2f}"):
            return self.query_over_temp()

        return False

    # --------------------------------------------------------------------------
    #   set_setpoint_preset
    # --------------------------------------------------------------------------

    def set_setpoint_preset(self, n: int) -> bool:
        """Instruct the Julabo to select another setpoint preset.

        Args:
          n (:obj:`int`): Setpoint to be used, either 1, 2 or 3.

        Returns: True if successful, False otherwise.
        """

        if not n in (1, 2, 3):
            pft(
                "WARNING: Received illegal setpoint preset.\n"
                "Must be either 1, 2 or 3."
            )
            return False

        if self.write_(f"OUT_MODE_01 {(n - 1):d}"):
            self.state.setpoint_preset = n
            return True

        return False

    # --------------------------------------------------------------------------
    #   set_sendpoint
    # --------------------------------------------------------------------------

    def set_setpoint(self, value: float) -> bool:
        """Set the temperature setpoint #1, #2 or #3, depending on which one is
        currently the active preset. Subsequently, the Julabo is queried for the
        obtained value, which might be different than the one requested.

        Returns: True if all communication was successful, False otherwise.
        """

        success = False

        if self.state.setpoint_preset == 1:
            success = self.set_setpoint_1(value)
            self.state.setpoint = self.state.setpoint_1
        elif self.state.setpoint_preset == 2:
            success = self.set_setpoint_2(value)
            self.state.setpoint = self.state.setpoint_2
        elif self.state.setpoint_preset == 3:
            success = self.set_setpoint_3(value)
            self.state.setpoint = self.state.setpoint_3

        return success

    # --------------------------------------------------------------------------
    #   set_sendpoint_1
    # --------------------------------------------------------------------------

    def set_setpoint_1(self, value: float) -> bool:
        """Set the temperature setpoint #1. Subsequently, the Julabo is queried
        for the obtained value, which might be different than the one requested.

        Returns: True if all communication was successful, False otherwise.
        """

        try:
            value = float(value)
        except (TypeError, ValueError) as err:
            pft(err)
            return False

        if self.write_(f"OUT_SP_00 {value:.2f}"):
            return self.query_setpoint_1()

        return False

    # --------------------------------------------------------------------------
    #   set_sendpoint_2
    # --------------------------------------------------------------------------

    def set_setpoint_2(self, value: float) -> bool:
        """Set the temperature setpoint #2. Subsequently, the Julabo is queried
        for the obtained value, which might be different than the one requested.

        Returns: True if all communication was successful, False otherwise.
        """

        try:
            value = float(value)
        except (TypeError, ValueError) as err:
            pft(err)
            return False

        if self.write_(f"OUT_SP_01 {value:.2f}"):
            return self.query_setpoint_2()

        return False

    # --------------------------------------------------------------------------
    #   set_sendpoint_3
    # --------------------------------------------------------------------------

    def set_setpoint_3(self, value: float) -> bool:
        """Set the temperature setpoint #3. Subsequently, the Julabo is queried
        for the obtained value, which might be different than the one requested.

        Returns: True if all communication was successful, False otherwise.
        """

        try:
            value = float(value)
        except (TypeError, ValueError) as err:
            pft(err)
            return False

        if self.write_(f"OUT_SP_02 {value:.2f}"):
            return self.query_setpoint_3()

        return False

    # --------------------------------------------------------------------------
    #   query_version
    # --------------------------------------------------------------------------

    def query_version(self) -> bool:
        """Query the version of the Julabo firmware and store it in the class
        member 'state'. Will be set to numpy.nan if unsuccessful.

        Returns: True if successful, False otherwise.
        """
        _success, reply = self.query_("VERSION")
        if isinstance(reply, str):
            self.state.version = reply
            return True

        self.state.version = ""
        return False

    # --------------------------------------------------------------------------
    #   query_status
    # --------------------------------------------------------------------------

    def query_status(self) -> bool:
        """Query the status or error message of the Julabo and store it in the
        class member 'state'. Will be set to numpy.nan if unsuccessful.

        Returns: True if successful, False otherwise.
        """
        _success, reply = self.query_("STATUS")
        if isinstance(reply, str):
            self.state.status = reply

            try:
                status_number = int(self.state.status[:3])
            except (TypeError, ValueError) as err:
                self.state.has_error = np.nan
                pft(err)
            else:
                if status_number < 0:
                    self.state.has_error = True
                else:
                    self.state.has_error = False

            return True

        self.state.status = np.nan
        self.state.has_error = np.nan
        return False

    # --------------------------------------------------------------------------
    #   query_temp_unit
    # --------------------------------------------------------------------------

    def query_temp_unit(self) -> bool:
        """Query the temperature unit used by the Julabo and store it in the
        class member 'state'. Will be set to "" if unsuccessful, else either "C"
        or "F".

        Returns: True if successful, False otherwise.
        """
        _success, reply = self.query_("IN_SP_06")
        if isinstance(reply, str):
            try:
                num = int(reply)
            except (TypeError, ValueError) as err:
                pft(err)
            else:
                self.state.temp_unit = "C" if num == 0 else "F"
                return True

        self.state.temp_unit = ""
        return False

    # --------------------------------------------------------------------------
    #   query_running
    # --------------------------------------------------------------------------

    def query_running(self) -> bool:
        """Query if the Julabo is running and store it in the class member
        'state'. Will be set to numpy.nan if unsuccessful.

        Returns: True if successful, False otherwise.
        """
        _success, reply = self.query_("IN_MODE_05")
        if isinstance(reply, str):
            try:
                ans = bool(int(reply))
            except (TypeError, ValueError) as err:
                pft(err)
            else:
                self.state.running = ans
                return True

        self.state.running = np.nan
        return False

    # --------------------------------------------------------------------------
    #   query_sub_temp
    # --------------------------------------------------------------------------

    def query_sub_temp(self) -> bool:
        """Query the low-temperature warning limit and store it in the class
        member 'state'. Will be set to numpy.nan if unsuccessful.

        Returns: True if successful, False otherwise.
        """
        _success, reply = self.query_("IN_SP_04")
        if isinstance(reply, str):
            try:
                num = float(reply)
            except (TypeError, ValueError) as err:
                pft(err)
            else:
                self.state.sub_temp = num
                return True

        self.state.sub_temp = np.nan
        return False

    # --------------------------------------------------------------------------
    #   query_over_temp
    # --------------------------------------------------------------------------

    def query_over_temp(self) -> bool:
        """Query the high-temperature warning limit and store it in the class
        member 'state'. Will be set to numpy.nan if unsuccessful.

        Returns: True if successful, False otherwise.
        """
        _success, reply = self.query_("IN_SP_03")
        if isinstance(reply, str):
            try:
                num = float(reply)
            except (TypeError, ValueError) as err:
                pft(err)
            else:
                self.state.over_temp = num
                return True

        self.state.over_temp = np.nan
        return False

    # --------------------------------------------------------------------------
    #   query_safe_temp
    # --------------------------------------------------------------------------

    def query_safe_temp(self) -> bool:
        """Query the screw-set excess temperature protection and store it in the
        class member 'state'. Will be set to numpy.nan if unsuccessful.

        Returns: True if successful, False otherwise.
        """
        _success, reply = self.query_("IN_PV_04")
        if isinstance(reply, str):
            try:
                num = float(reply)
            except (TypeError, ValueError) as err:
                pft(err)
            else:
                self.state.safe_temp = num
                return True

        self.state.safe_temp = np.nan
        return False

    # --------------------------------------------------------------------------
    #   query_safe_sens
    # --------------------------------------------------------------------------

    def query_safe_sens(self) -> bool:
        """Query the safety sensor temperature and store it in the class member
        'state'. Will be set to numpy.nan if unsuccessful.

        Returns: True if successful, False otherwise.
        """
        _success, reply = self.query_("IN_PV_03")
        if isinstance(reply, str):
            try:
                num = float(reply)
            except (TypeError, ValueError) as err:
                pft(err)
            else:
                self.state.safe_sens = num
                return True

        self.state.safe_sens = np.nan
        return False

    # --------------------------------------------------------------------------
    #   query_setpoint_preset
    # --------------------------------------------------------------------------

    def query_setpoint_preset(self) -> bool:
        """Query the setpoint preset currently used by the Julabo (#1, #2 or #3)
        and store it in the class member 'state'. Will be set to numpy.nan
        if unsuccessful.

        Returns: True if successful, False otherwise.
        """
        _success, reply = self.query_("IN_MODE_01")
        if isinstance(reply, str):
            try:
                num = int(reply)
            except (TypeError, ValueError) as err:
                pft(err)
            else:
                self.state.setpoint_preset = num + 1
                return True

        self.state.setpoint_preset = np.nan
        return False

    # --------------------------------------------------------------------------
    #   query_setpoint
    # --------------------------------------------------------------------------

    def query_setpoint(self) -> bool:
        """Query the temperature setpoint #1, #2 or #3, depending on which one
        is currently the active preset, and store it in the class member
        'state'. Will be set to numpy.nan if unsuccessful.

        Returns: True if successful, False otherwise.
        """
        success = False

        if self.state.setpoint_preset == 1:
            success = self.query_setpoint_1()
            self.state.setpoint = self.state.setpoint_1
        elif self.state.setpoint_preset == 2:
            success = self.query_setpoint_2()
            self.state.setpoint = self.state.setpoint_2
        elif self.state.setpoint_preset == 3:
            success = self.query_setpoint_3()
            self.state.setpoint = self.state.setpoint_3

        return success

    # --------------------------------------------------------------------------
    #   query_setpoint_1
    # --------------------------------------------------------------------------

    def query_setpoint_1(self) -> bool:
        """Query the temperature setpoint #1 and store it in the class member
        'state'. Will be set to numpy.nan if unsuccessful.

        Returns: True if successful, False otherwise.
        """
        _success, reply = self.query_("IN_SP_00")
        if isinstance(reply, str):
            try:
                num = float(reply)
            except (TypeError, ValueError) as err:
                pft(err)
            else:
                self.state.setpoint_1 = num
                return True

        self.state.setpoint_1 = np.nan
        return False

    # --------------------------------------------------------------------------
    #   query_setpoint_2
    # --------------------------------------------------------------------------

    def query_setpoint_2(self) -> bool:
        """Query the temperature setpoint #2 and store it in the class member
        'state'. Will be set to numpy.nan if unsuccessful.

        Returns: True if successful, False otherwise.
        """
        _success, reply = self.query_("IN_SP_01")
        if isinstance(reply, str):
            try:
                num = float(reply)
            except (TypeError, ValueError) as err:
                pft(err)
            else:
                self.state.setpoint_2 = num
                return True

        self.state.setpoint_2 = np.nan
        return False

    # --------------------------------------------------------------------------
    #   query_setpoint_3
    # --------------------------------------------------------------------------

    def query_setpoint_3(self) -> bool:
        """Query the temperature setpoint #3 and store it in the class member
        'state'. Will be set to numpy.nan if unsuccessful.

        Returns: True if successful, False otherwise.
        """
        _success, reply = self.query_("IN_SP_02")
        if isinstance(reply, str):
            try:
                num = float(reply)
            except (TypeError, ValueError) as err:
                pft(err)
            else:
                self.state.setpoint_3 = num
                return True

        self.state.setpoint_3 = np.nan
        return False

    # --------------------------------------------------------------------------
    #   query_bath_temp
    # --------------------------------------------------------------------------

    def query_bath_temp(self) -> bool:
        """Query the current bath temperature and store it in the class member
        'state'. Will be set to numpy.nan if unsuccessful.

        Returns: True if successful, False otherwise.
        """
        _success, reply = self.query_("IN_PV_00")
        if isinstance(reply, str):
            try:
                num = float(reply)
            except (TypeError, ValueError) as err:
                pft(err)
            else:
                self.state.bath_temp = num
                return True

        self.state.bath_temp = np.nan
        return False

    # --------------------------------------------------------------------------
    #   query_pt100_temp
    # --------------------------------------------------------------------------

    def query_pt100_temp(self) -> bool:
        """Query the current external Pt100 temperature sensor and store it in
        the class member 'state'. Will be set to numpy.nan if no external sensor
        is connected or when communication is unsuccessful.

        Returns: True if successful, False otherwise.
        """
        _success, reply = self.query_("IN_PV_02")
        if isinstance(reply, str):
            if reply == "---.--":  # Not connected
                self.state.pt100_temp = np.nan
                return True

            try:
                num = float(reply)
            except (TypeError, ValueError) as err:
                pft(err)
            else:
                self.state.pt100_temp = num
                return True

        self.state.pt100_temp = np.nan
        return False

    # --------------------------------------------------------------------------
    #   query_common_readings
    # --------------------------------------------------------------------------

    def query_common_readings(self) -> bool:
        """Query the most common readings:
        - Running?
        - Setpoint
        - Bath temperature
        - Pt100 temperature
        - Safe sensor temperature
        - Status

        Returns: True if successful, False otherwise.
        """

        success = True
        success &= self.query_running()
        success &= self.query_setpoint()
        success &= self.query_bath_temp()
        success &= self.query_pt100_temp()
        success &= self.query_safe_sens()
        success &= self.query_status()

        return success

    # --------------------------------------------------------------------------
    #   report
    # --------------------------------------------------------------------------

    def report(self, update_readings: bool = True):
        """Print info to the terminal, useful for debugging"""

        C = self.state  # Shorthand notation
        w1 = 10  # Label width
        w2 = 8  # Value width

        # Update readings
        if update_readings:
            self.query_common_readings()

        # Print to terminal
        print(self.state.version)
        print(f"{'Temp. unit':<{w1}s}: {C.temp_unit:<{w2}s}", end="")
        print(f"{'Sel. setp.':<{w1}s}: {C.setpoint_preset:<{w2}.0f}")
        print(f"{'Sub temp.':<{w1}s}: {C.sub_temp:<{w2}.2f}", end="")
        print(f"{'Over temp.':<{w1}s}: {C.over_temp:<{w2}.2f}")
        print(f"{'':<{w1}s}  {'':<{w2}s}", end="")
        print(f"{'Safe temp.':<{w1}s}: {C.safe_temp:<{w2}.2f}")
        print()
        if not isinstance(C.running, bool):
            print("COMMUNICATION ERROR")
        else:
            print("--> RUNNING <--" if C.running else "IDLE")
        print(f"{'Setpoint':<{w1}s}: {C.setpoint:<{w2}.2f}", end="")
        print(f"{'Bath temp.':<{w1}s}: { C.bath_temp:<{w2}.2f}")
        print(f"{'Safe sens':<{w1}s}: {C.safe_sens:<{w2}.2f}", end="")
        print(f"{'Pt100':<{w1}s}: {C.pt100_temp:<{w2}.2f}")
        print()
        print(f"Status msg: {C.status}")

    # --------------------------------------------------------------------------
    #   query_
    # --------------------------------------------------------------------------

    def query_(self, *args, **kwargs) -> Tuple[bool, Union[str, bytes, None]]:
        """Wrapper for :meth:`dvg_qdevices.query` to add enforcing of time gaps
        between commands as per the Julabo manual.

        Returns:
            :obj:`tuple`:
                - success (:obj:`bool`):
                    True if successful, False otherwise.

                - reply (:obj:`str` | :obj:`bytes` | :obj:`None`):
                    Reply received from the device, either as ASCII string
                    (default) or as bytes when ``returns_ascii`` was set to
                    :const:`False`. :obj:`None` if unsuccessful.
        """

        # fmt: off
        while (
            (time.perf_counter() - self.state.t_prev_in < DELAY_COMMAND_IN) or
            (time.perf_counter() - self.state.t_prev_out < DELAY_COMMAND_OUT)
        ):
            pass
        # fmt: on

        success, reply = super().query(*args, **kwargs)
        self.state.t_prev_in = time.perf_counter()

        return success, reply

    # --------------------------------------------------------------------------
    #   write_
    # --------------------------------------------------------------------------

    def write_(self, *args, **kwargs) -> bool:
        """Wrapper for :meth:`dvg_qdevices.write` to add enforcing of time gaps
        between commands as per the Julabo manual.

        Returns: True if successful, False otherwise.
        """

        # fmt: off
        while (
            (time.perf_counter() - self.state.t_prev_in < DELAY_COMMAND_IN) or
            (time.perf_counter() - self.state.t_prev_out < DELAY_COMMAND_OUT)
        ):
            pass
        # fmt: on

        success = super().write(*args, **kwargs)
        self.state.t_prev_out = time.perf_counter()

        return success


# -----------------------------------------------------------------------------
#   Main: Will show a demo when run from the terminal
# -----------------------------------------------------------------------------

if __name__ == "__main__":
    import os
    import sys

    # Path to the config textfile containing the (last used) RS232 port
    PATH_CONFIG = "config/port_Julabo_circulator.txt"

    # Create connection to Julabo over RS232
    julabo = Julabo_circulator()
    if julabo.auto_connect(filepath_last_known_port=PATH_CONFIG):
        julabo.begin()  # Retrieve necessary parameters
    else:
        time.sleep(1)
        sys.exit(0)

    if os.name == "nt":
        import msvcrt

        running_Windows = True
    else:
        running_Windows = False

    # Prepare
    send_setpoint = 22.0
    do_send_setpoint = False

    # Loop
    done = False
    while not done:
        # Check if a new setpoint has to be send
        if do_send_setpoint:
            julabo.set_setpoint(send_setpoint)
            do_send_setpoint = False

        # Measure and report
        julabo.query_common_readings()

        if running_Windows:
            os.system("cls")
            julabo.report(update_readings=False)

            print("\nPress Q to quit.")
            print("Press S to enter new setpoint.")
            print("Press O to toggle the Julabo on/off.")
        else:
            os.system("clear")
            julabo.report(update_readings=False)

            print("\nPress Control + C to quit.")
            print("No other keyboard input possible because OS is not Windows.")

        sys.stdout.flush()

        # Process keyboard input
        if running_Windows:
            if msvcrt.kbhit():
                key = msvcrt.getch()

                if key in (b"q", b"Q"):
                    print("\nAre you sure you want to quit [y/n]?")
                    if msvcrt.getch() == b"y":
                        print("Switching off Julabo and quitting.")
                        done = True

                elif key in (b"s", b"S"):
                    user_input = input("\nEnter new setpoint: ")

                    try:
                        send_setpoint = float(user_input)
                    except Exception:
                        print("Error: Could not parse float value.")
                    else:
                        do_send_setpoint = True

                elif key in (b"o", b"O"):
                    if julabo.state.running:
                        julabo.turn_off()
                    else:
                        julabo.turn_on()

        # Slow down update period
        time.sleep(0.5)

    julabo.turn_off()
    time.sleep(1)  # Give time to turn off

    julabo.close()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Multithreaded PyQt/PySide GUI to interface with a Keysight (former HP or
Agilent) 34970A/34972A data acquisition/switch unit.
"""
__author__ = "Dennis van Gils"
__authoremail__ = "vangils.dennis@gmail.com"
__url__ = "https://github.com/Dennis-van-Gils/python-dvg-devices"
__date__ = "23-05-2024"
__version__ = "1.5.0"
print(__url__)
# pylint: disable=wrong-import-position, missing-function-docstring, bare-except

import sys
import time

import qtpy
from qtpy import QtCore, QtGui, QtWidgets as QtWid
from qtpy.QtCore import Slot  # type: ignore

import pyvisa
import matplotlib.pyplot as plt
import numpy as np
import pyqtgraph as pg

from dvg_pyqtgraph_threadsafe import (
    ThreadSafeCurve,
    HistoryChartCurve,
    LegendSelect,
    PlotManager,
)
import dvg_pyqt_controls as controls
from dvg_pyqt_filelogger import FileLogger
from dvg_devices.Keysight_3497xA_protocol_SCPI import Keysight_3497xA
from dvg_devices.Keysight_3497xA_qdev import Keysight_3497xA_qdev, INFINITY_CAP

TRY_USING_OPENGL = True
if TRY_USING_OPENGL:
    try:
        import OpenGL.GL as gl  # pylint: disable=unused-import
    except:  # pylint: disable=bare-except
        print("OpenGL acceleration: Disabled")
        print("To install: `conda install pyopengl` or `pip install pyopengl`")
    else:
        print("OpenGL acceleration: Enabled")
        pg.setConfigOptions(useOpenGL=True)
        pg.setConfigOptions(antialias=True)
        pg.setConfigOptions(enableExperimental=True)

# Global pyqtgraph configuration
# pg.setConfigOptions(leftButtonPan=False)
pg.setConfigOption("foreground", "#EEE")

# VISA address of the Keysight 3497xA data acquisition/switch unit containing a
# multiplexer plug-in module. Hence, we simply call this device a 'mux'.
# MUX_VISA_ADDRESS = "USB0::0x0957::0x2007::MY49018071::INSTR"
MUX_VISA_ADDRESS = "GPIB0::9::INSTR"

# SCPI commands to be send to the mux to set up the scan cycle.
"""
scan_list = "(@101:110)"
MUX_SCPI_COMMANDS = [
            f"rout:open {scan_list}",
            f"conf:temp TC,J,{scan_list}",
            f"unit:temp C,{scan_list}",
            f"sens:temp:tran:tc:rjun:type INT,{scan_list}",
            f"sens:temp:tran:tc:check ON,{scan_list}",
            f"sens:temp:nplc 1,{scan_list}",
            f"rout:scan {scan_list}",
]
"""

scan_list = "(@101:110)"
MUX_SCPI_COMMANDS = [
    f"rout:open {scan_list}",
    f"conf:res 1e6,{scan_list}",
    f"sens:res:nplc 1,{scan_list}",
    f"rout:scan {scan_list}",
]

# fmt: off
DAQ_INTERVAL_MS   = 1000  # [ms] Update interval for the mux to perform a scan
CHART_INTERVAL_MS = 1000  # [ms] Update interval for all charts
CHART_CAPACITY    = 1800  # [samples]
# fmt: on

# Show debug info in terminal? Warning: Slow! Do not leave on unintentionally.
DEBUG = False

# ------------------------------------------------------------------------------
#   MainWindow
# ------------------------------------------------------------------------------


class MainWindow(QtWid.QWidget):
    def __init__(
        self,
        qdev: Keysight_3497xA_qdev,
        qlog: FileLogger,
        parent=None,
        **kwargs,
    ):
        super().__init__(parent, **kwargs)

        self.qdev = qdev
        self.qlog = qlog

        self.setWindowTitle("Keysight 3497xA control")
        self.setGeometry(40, 60, 1200, 0)
        self.setFont(QtGui.QFont("Arial", 9))
        self.setStyleSheet(
            controls.SS_TEXTBOX_READ_ONLY
            + controls.SS_GROUP
            + controls.SS_HOVER
        )

        self.timer_GUI = QtCore.QTimer()
        self.timer_GUI.timeout.connect(self.update_GUI)

        # ----------------------------------------------------------------------
        #   Top grid
        # ----------------------------------------------------------------------

        self.qlbl_title = QtWid.QLabel("Keysight 3497xA control")
        self.qlbl_title.setFont(
            QtGui.QFont("Palatino", 14, weight=QtGui.QFont.Weight.Bold)
        )
        self.qlbl_cur_date_time = QtWid.QLabel("00-00-0000    00:00:00")
        self.qpbt_record = controls.create_Toggle_button(
            "Click to start recording to file", minimumHeight=40
        )
        self.qpbt_record.setMinimumWidth(400)
        # fmt: off
        self.qpbt_record.clicked.connect(lambda state: qlog.record(state))  # pylint: disable=unnecessary-lambda
        # fmt: on

        self.qpbt_exit = QtWid.QPushButton("Exit")
        self.qpbt_exit.clicked.connect(self.close)
        self.qpbt_exit.setMinimumHeight(30)

        grid_top = QtWid.QGridLayout()
        # fmt: off
        grid_top.addWidget(self.qlbl_title        , 0, 0, QtCore.Qt.AlignmentFlag.AlignCenter)
        grid_top.addWidget(self.qpbt_exit         , 0, 2, QtCore.Qt.AlignmentFlag.AlignRight)
        grid_top.addWidget(self.qlbl_cur_date_time, 1, 0, QtCore.Qt.AlignmentFlag.AlignCenter)
        grid_top.addWidget(self.qpbt_record       , 2, 0, QtCore.Qt.AlignmentFlag.AlignCenter)
        # fmt: on
        grid_top.setColumnMinimumWidth(0, 420)
        grid_top.setColumnStretch(1, 1)

        # ----------------------------------------------------------------------
        #   Chart: Mux readings
        # ----------------------------------------------------------------------

        # GraphicsLayoutWidget
        self.gw_mux = pg.GraphicsLayoutWidget()

        p = {"color": "#EEE", "font-size": "12pt"}
        self.pi_mux = self.gw_mux.addPlot()
        self.pi_mux.setClipToView(True)
        self.pi_mux.showGrid(x=1, y=1)
        self.pi_mux.setTitle("Mux readings", **p)
        self.pi_mux.setLabel("bottom", "history (min)", **p)
        self.pi_mux.setLabel("left", "misc. units", **p)
        self.pi_mux.setMenuEnabled(True)
        self.pi_mux.enableAutoRange(axis=pg.ViewBox.XAxis, enable=False)
        self.pi_mux.enableAutoRange(axis=pg.ViewBox.YAxis, enable=True)
        self.pi_mux.setAutoVisible(y=True)

        # ----------------------------------------------------------------------
        #   Create history charts depending on the number of scan channels
        # ----------------------------------------------------------------------

        self.tscurves_mux: list[ThreadSafeCurve] = []

        cm = plt.get_cmap("gist_rainbow")
        for idx, channel in enumerate(qdev.dev.state.all_scan_list_channels):
            color = cm(1.0 * idx / qdev.dev.state.N_channels)  # RGBA tuple
            color = np.array(color) * 255
            pen = pg.mkPen(color=color, width=2)

            self.tscurves_mux.append(
                HistoryChartCurve(
                    capacity=CHART_CAPACITY,
                    linked_curve=self.pi_mux.plot(pen=pen, name=channel),
                )
            )

        # ----------------------------------------------------------------------
        #   Legend
        # ----------------------------------------------------------------------

        legend = LegendSelect(linked_curves=self.tscurves_mux)
        legend.grid.setVerticalSpacing(0)

        self.qgrp_legend = QtWid.QGroupBox("Legend")
        self.qgrp_legend.setLayout(legend.grid)

        # ----------------------------------------------------------------------
        #   PlotManager
        # ----------------------------------------------------------------------

        self.plot_manager = PlotManager(parent=self)
        self.plot_manager.add_autorange_buttons(linked_plots=self.pi_mux)
        self.plot_manager.add_preset_buttons(
            linked_plots=self.pi_mux,
            linked_curves=self.tscurves_mux,
            presets=[
                {
                    "button_label": "0:30",
                    "x_axis_label": "history (sec)",
                    "x_axis_divisor": 1,
                    "x_axis_range": (-30, 0),
                },
                {
                    "button_label": "01:00",
                    "x_axis_label": "history (sec)",
                    "x_axis_divisor": 1,
                    "x_axis_range": (-60, 0),
                },
                {
                    "button_label": "03:00",
                    "x_axis_label": "history (min)",
                    "x_axis_divisor": 60,
                    "x_axis_range": (-3, 0),
                },
                {
                    "button_label": "05:00",
                    "x_axis_label": "history (min)",
                    "x_axis_divisor": 60,
                    "x_axis_range": (-5, 0),
                },
                {
                    "button_label": "10:00",
                    "x_axis_label": "history (min)",
                    "x_axis_divisor": 60,
                    "x_axis_range": (-10, 0),
                },
                {
                    "button_label": "30:00",
                    "x_axis_label": "history (min)",
                    "x_axis_divisor": 60,
                    "x_axis_range": (-30, 0),
                },
            ],
        )
        self.plot_manager.add_clear_button(linked_curves=self.tscurves_mux)

        qgrp_history = QtWid.QGroupBox("History")
        qgrp_history.setLayout(self.plot_manager.grid)

        # ----------------------------------------------------------------------
        #   Right-panel grid
        # ----------------------------------------------------------------------

        vbox1 = QtWid.QVBoxLayout()
        vbox1.addWidget(
            self.qgrp_legend,
            stretch=0,
            alignment=QtCore.Qt.AlignmentFlag.AlignTop,
        )
        vbox1.addWidget(
            qgrp_history,
            stretch=0,
            alignment=QtCore.Qt.AlignmentFlag.AlignTop,
        )
        vbox1.addStretch(1)

        # ----------------------------------------------------------------------
        #   Round up full window
        # ----------------------------------------------------------------------

        hbox1 = QtWid.QHBoxLayout()
        hbox1.addWidget(
            qdev.qgrp,
            stretch=0,
            alignment=QtCore.Qt.AlignmentFlag.AlignTop,
        )
        hbox1.addWidget(self.gw_mux, stretch=1)
        hbox1.addLayout(vbox1)

        vbox = QtWid.QVBoxLayout(self)
        vbox.addLayout(grid_top)
        vbox.addLayout(hbox1)
        vbox.addStretch(1)

    # --------------------------------------------------------------------------
    #   Handle controls
    # --------------------------------------------------------------------------

    @Slot()
    def update_GUI(self):
        cur_date_time = QtCore.QDateTime.currentDateTime()
        self.qlbl_cur_date_time.setText(
            cur_date_time.toString("dd-MM-yyyy")
            + "    "
            + cur_date_time.toString("HH:mm:ss")
        )

        # Update curves
        for tscurve in self.tscurves_mux:
            tscurve.update()


# ------------------------------------------------------------------------------
#   Main
# ------------------------------------------------------------------------------

if __name__ == "__main__":

    # --------------------------------------------------------------------------
    #   Connect to Keysight 3497xA (mux)
    # --------------------------------------------------------------------------

    rm = pyvisa.ResourceManager()
    mux = Keysight_3497xA(MUX_VISA_ADDRESS, "MUX")

    try:
        if mux.connect(rm):
            mux.begin(MUX_SCPI_COMMANDS)
    except ValueError as e:
        # No connection could be made to the VISA device because module
        # dependencies are missing. Print error, not raise and continue to
        # show the GUI.
        print(e)

    # --------------------------------------------------------------------------
    #   Create application
    # --------------------------------------------------------------------------

    main_thread = QtCore.QThread.currentThread()
    if isinstance(main_thread, QtCore.QThread):
        main_thread.setObjectName("MAIN")  # For DEBUG info

    if qtpy.PYQT6 or qtpy.PYSIDE6:
        sys.argv += ["-platform", "windows:darkmode=0"]
    app = QtWid.QApplication(sys.argv)
    app.setStyle("Fusion")

    # --------------------------------------------------------------------------
    #   Set up communication threads for the mux
    # --------------------------------------------------------------------------

    def postprocess_mux_fun():
        """Will be called during an 'worker_DAQ' update, after a mux scan has
        been performed. We use it to parse out the scan readings into separate
        variables and log it to file.
        """
        cur_date_time = QtCore.QDateTime.currentDateTime()
        now = time.perf_counter()

        # DEBUG info
        # dprint(f"thread: {QtCore.QThread.currentThread().objectName()}")

        if mux_qdev.is_MUX_scanning:
            readings = mux.state.readings
            for idx in range(mux.state.N_channels):
                if readings[idx] > INFINITY_CAP:
                    readings[idx] = np.nan
        else:
            readings = [np.nan] * mux.state.N_channels
            mux.state.readings = readings

        # Add readings to charts
        for idx, tscurve in enumerate(window.tscurves_mux):
            tscurve.appendData(now, readings[idx])

        # Log data to file
        log.update(
            filepath=f"data_mux_{cur_date_time.toString('yyMMdd_HHmmss')}.txt",
            mode="w",
        )

    mux_qdev = Keysight_3497xA_qdev(
        dev=mux,
        DAQ_interval_ms=DAQ_INTERVAL_MS,
        DAQ_postprocess_MUX_scan_function=postprocess_mux_fun,
        debug=DEBUG,
    )
    # mux_qdev.set_table_readings_format(".5e")
    mux_qdev.start()

    # --------------------------------------------------------------------------
    #   File logger
    # --------------------------------------------------------------------------

    def write_header_to_log():
        ch_labels = [f"CH{ch}" for ch in mux.state.all_scan_list_channels]
        log.write("time [s]\t")
        log.write(f"{chr(9).join(ch_labels)}")  # [TAB]-delimited
        log.write("\n")

    def write_data_to_log():
        log.write(f"{log.elapsed():.3f}")
        for idx, _ch in enumerate(mux.state.all_scan_list_channels):
            if len(mux.state.readings) <= idx:
                log.write(f"\t{np.nan:.5e}")
            else:
                log.write(f"\t{mux.state.readings[idx]:.5e}")
        log.write("\n")

    log = FileLogger(
        write_header_function=write_header_to_log,
        write_data_function=write_data_to_log,
    )
    log.signal_recording_started.connect(
        lambda filepath: window.qpbt_record.setText(
            f"Recording to file: {filepath}"
        )
    )
    log.signal_recording_stopped.connect(
        lambda: window.qpbt_record.setText("Click to start recording to file")
    )

    # --------------------------------------------------------------------------
    #   Start the main GUI event loop
    # --------------------------------------------------------------------------

    def about_to_quit():
        print("About to quit")
        app.processEvents()
        mux_qdev.quit()
        try:
            mux.close()
        except:
            pass
        try:
            rm.close()
        except:
            pass

    app.aboutToQuit.connect(about_to_quit)

    window = MainWindow(qdev=mux_qdev, qlog=log)
    window.plot_manager.perform_preset(2)  # Init time axis of the history chart
    window.timer_GUI.start(CHART_INTERVAL_MS)
    window.show()

    sys.exit(app.exec())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""RS232 function library for Bronkhorst mass flow controllers (MFC) using the
FLOW-BUS protocol.

Only the ASCII version is supported, not the enhanced binary version. This
library sends and receives messages to/from all nodes (code 80), hence just one
MFC is assumed per port.

When this module is directly run from the terminal a demo will be shown.
"""
__author__ = "Dennis van Gils"
__authoremail__ = "vangils.dennis@gmail.com"
__url__ = "https://github.com/Dennis-van-Gils/python-dvg-devices"
__date__ = "23-05-2024"
__version__ = "1.5.0"
# pylint: disable=missing-function-docstring

import sys
import struct
from typing import Union, Tuple

import numpy as np

from dvg_devices.BaseDevice import SerialDevice


class Bronkhorst_MFC(SerialDevice):
    class State:
        """Container for the process and measurement variables"""

        setpoint: float = np.nan
        """Setpoint read out of the MFC  [ln/min]"""
        flow_rate: float = np.nan
        """Flow rate measured by the MFC [ln/min]"""

    def __init__(
        self,
        name="MFC",
        long_name="Bronkhorst MFC",
        connect_to_serial_number=None,
    ):
        super().__init__(name=name, long_name=long_name)

        # Default serial settings
        self.serial_settings = {
            "baudrate": 38400,
            "timeout": 0.1,
            "write_timeout": 0.1,
        }
        self.set_read_termination("\r\n")
        self.set_write_termination("\r\n")

        self.set_ID_validation_query(
            ID_validation_query=self.ID_validation_query,
            valid_ID_broad="8002716300",
            valid_ID_specific=connect_to_serial_number,
        )

        # Container for the process and measurement variables
        self.state = self.State()

        # fmt: off
        self.serial_str: str = ""  # Serial number of the MFC
        self.model_str : str = ""  # Model of the MFC
        self.fluid_name: str = ""  # Fluid for which the MFC is calibrated
        self.max_flow_rate: float = np.nan  # Max. capacity [ln/min]
        # fmt: on

    # --------------------------------------------------------------------------
    #   ID_validation_query
    # --------------------------------------------------------------------------

    def ID_validation_query(self) -> Tuple[str, Union[str, None]]:
        _success, reply = self.query(":0780047163716300")
        if isinstance(reply, str):
            broad_reply = reply[3:13]  # Expected: "8002716300"
            specific_reply = bytearray.fromhex(reply[13:-2]).decode()  # Serial
        else:
            broad_reply = ""
            specific_reply = None

        return broad_reply, specific_reply

    # --------------------------------------------------------------------------
    #   begin
    # --------------------------------------------------------------------------

    def begin(self):
        """This function should run directly after having established a
        connection to the device.

        Query the serial number, model, fluid and the maximum mass flow rate of
        the MFC and store these in the class member 'state's. The max flow rate
        is mandatory to be known, because it is used to set and read the
        setpoint, and to read the flow rate.

        Returns: True if successful, False otherwise.
        """
        success = self.query_serial_str()
        success &= self.query_model_str()
        success &= self.query_fluid_name()
        success &= self.query_max_flow_rate()
        success &= self.query_setpoint()
        success &= self.query_flow_rate()
        return success

    # --------------------------------------------------------------------------
    #   query_serial_str
    # --------------------------------------------------------------------------

    def query_serial_str(self) -> bool:
        """Query the serial number and store it in the class member 'serial_str'

        Returns: True if successful, False otherwise.
        """
        _success, reply = self.query(":0780047163716300")
        if isinstance(reply, str) and reply[3:13] == "8002716300":
            self.serial_str = bytearray.fromhex(reply[13:-2]).decode()
            return True

        self.serial_str = ""
        return False

    # --------------------------------------------------------------------------
    #   query_model_str
    # --------------------------------------------------------------------------

    def query_model_str(self) -> bool:
        """Query the model name of the MFC and store it in the class member
        'state'.

        Returns: True if successful, False otherwise.
        """
        _success, reply = self.query(":0780047162716200")
        if isinstance(reply, str):
            self.model_str = bytearray.fromhex(reply[13:-2]).decode()
            return True

        self.model_str = ""
        return False

    # --------------------------------------------------------------------------
    #   query_fluid_name
    # --------------------------------------------------------------------------

    def query_fluid_name(self) -> bool:
        """Query the fluid name that the MFC is calibrated for and store it in
        the class member 'state'.

        Returns: True if successful, False otherwise.
        """
        _success, reply = self.query(":078004017101710A")
        if isinstance(reply, str):
            self.fluid_name = bytearray.fromhex(reply[13:-2]).decode()
            return True

        self.fluid_name = ""
        return False

    # --------------------------------------------------------------------------
    #   query_max_flow_rate
    # --------------------------------------------------------------------------

    def query_max_flow_rate(self) -> bool:
        """Query the maximum mass flow rate in [ln/min] of the MFC and store it
        in the class member 'state'. This value is mandatory to be known, because it is
        used to set and read the setpoint, and to read the flow rate.

        Returns: True if successful, False otherwise.
        """
        _success, reply = self.query(":068004014D014D")
        if isinstance(reply, str):
            self.max_flow_rate = hex_to_32bit_IEEE754_float(reply[11:])
            return True

        self.max_flow_rate = np.nan
        return False

    # --------------------------------------------------------------------------
    #   query_setpoint
    # --------------------------------------------------------------------------

    def query_setpoint(self) -> bool:
        """Query the mass flow rate setpoint in [ln/min] set at the MFC and
        store it in the class member 'state' 'state'. Will be set to None if
        unsuccessful.

        Returns: True if successful, False otherwise.
        """
        _success, reply = self.query(":06800401210121")
        if isinstance(reply, str):
            try:
                num = int(reply[-4:], 16)
            except ValueError:
                pass
            else:
                self.state.setpoint = num / 32000.0 * self.max_flow_rate
                return True

        self.state.setpoint = np.nan
        return False

    # --------------------------------------------------------------------------
    #   query_flow_rate
    # --------------------------------------------------------------------------

    def query_flow_rate(self) -> bool:
        """Query the mass flow rate in [ln/min] measured by the MFC and
        store it in the class member 'state'. Will be set to None if
        unsuccessful.

        Returns: True if successful, False otherwise.
        """
        _success, reply = self.query(":06800401210120")
        if isinstance(reply, str):
            try:
                num = int(reply[-4:], 16)
            except ValueError:
                pass
            else:
                self.state.flow_rate = num / 32000.0 * self.max_flow_rate
                return True

        self.state.flow_rate = np.nan
        return False

    # --------------------------------------------------------------------------
    #   send_setpoint
    # --------------------------------------------------------------------------

    def send_setpoint(self, setpoint: float) -> bool:
        """Send a new mass flow rate setpoint in [ln/min] to the MFC.

        Args:
            setpoint (float): mass flow rate in [ln/min].

        Returns: True if successful, False otherwise.
        """
        # Transform setpoint and limit
        setpoint = int(setpoint / self.max_flow_rate * 32000)
        setpoint = max(0, min(setpoint, 32000))

        _success, reply = self.query(f":0680010121{setpoint:04x}")
        if isinstance(reply, str) and reply[5:].strip() == "000005":
            return True  # Also checked status reply

        return False


# ------------------------------------------------------------------------------
#   hex_to_32bit_IEEE754_float
# ------------------------------------------------------------------------------


def hex_to_32bit_IEEE754_float(hex_str) -> float:
    """Transform a string containing a hexidecimal representation of a 32-bits
    IEEE754-formatted float value to a float
    """
    return (struct.unpack("f", struct.pack("i", int(hex_str, 16))))[0]


# ------------------------------------------------------------------------------
#   Main: Will show a demo when run from the terminal
# ------------------------------------------------------------------------------

if __name__ == "__main__":
    import os
    import time

    # Config file containing COM port address
    PATH_CONFIG = "config/port_Bronkhorst_MFC_1.txt"

    # Serial number of the Bronkhorst MFC to connect to.
    # Set to '' or None to connect to any Bronkhorst MFC.
    # SERIAL_MFC = "M16216843A"
    SERIAL_MFC = None

    # Create connection to Bronkhorst MFC over RS232
    mfc = Bronkhorst_MFC(connect_to_serial_number=SERIAL_MFC)
    if mfc.auto_connect(filepath_last_known_port=PATH_CONFIG):
        mfc.begin()  # Retrieve necessary parameters
        print(f"  Serial  : {mfc.serial_str}")
        print(f"  Model   : {mfc.model_str}")
        print(f"  Fluid   : {mfc.fluid_name}")
        print(f"  Capacity: {mfc.max_flow_rate:.2f} ln/min")
    else:
        time.sleep(1)
        sys.exit(0)

    if os.name == "nt":
        import msvcrt

        running_Windows = True
        print("\nPress Q to quit.")
        print("Press S to enter new setpoint.")
    else:
        running_Windows = False
        print("\nPress Control + C to quit.")
        print("No other keyboard input possible because OS is not Windows.")

    # Prepare
    send_setpoint = 0.0
    do_send_setpoint = False

    mfc.query_setpoint()
    print(f"\nSetpoint       : {mfc.state.setpoint:6.2f} ln/min")

    # Loop
    done = False
    while not done:
        # Check if a new setpoint has to be send
        if do_send_setpoint:
            mfc.send_setpoint(send_setpoint)
            mfc.query_setpoint()
            print(f"\nSetpoint       : {mfc.state.setpoint:6.2f} ln/min")
            do_send_setpoint = False

        # Measure and report the flow rate
        mfc.query_flow_rate()
        print(f"\rMeas. flow rate: {mfc.state.flow_rate:6.2f} ln/min", end="")
        sys.stdout.flush()

        # Process keyboard input
        if running_Windows:
            if msvcrt.kbhit():
                key = msvcrt.getch()
                if key == b"q":
                    print("\nAre you sure you want to quit [y/n]?")
                    if msvcrt.getch() == b"y":
                        print("Quitting.")
                        done = True
                    else:
                        do_send_setpoint = True  # Esthestics
                elif key == b"s":
                    input_str = input("\nEnter new setpoint [ln/min]: ")
                    try:
                        input_float = float(input_str)
                    except ValueError:
                        input_float = 0.0
                    send_setpoint = input_float
                    do_send_setpoint = True

        # Slow down update period
        time.sleep(0.02)

    mfc.close()
    time.sleep(1)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""PyQt/PySide module to provide multithreaded communication and periodical data
acquisition for a Compax3 traverse controller.
"""
__author__ = "Dennis van Gils"
__authoremail__ = "vangils.dennis@gmail.com"
__url__ = "https://github.com/Dennis-van-Gils/python-dvg-devices"
__date__ = "23-05-2024"
__version__ = "1.5.0"
# pylint: disable=missing-function-docstring, multiple-statements

from qtpy import QtCore, QtGui, QtWidgets as QtWid
from qtpy.QtCore import Slot  # type: ignore

import dvg_pyqt_controls as controls
from dvg_qdeviceio import QDeviceIO, DAQ_TRIGGER
from dvg_devices.Compax3_servo_protocol_RS232 import Compax3_servo


class Compax3_servo_qdev(QDeviceIO):
    """Manages multithreaded communication and periodical data acquisition for
    a Compax3 traverse controller, referred to as the 'device'.

    In addition, it also provides PyQt/PySide GUI objects for control of the
    device. These can be incorporated into your application.

    All device I/O operations will be offloaded to 'workers', each running in
    a newly created thread.

    (*): See 'dvg_qdeviceio.QDeviceIO()' for details.

    Args:
        dev:
            Reference to a
            'dvg_devices.Compax3_servo_protocol_RS232.Compax3_servo'
            instance.

        debug:
            Show debug info in terminal? Warning: Slow! Do not leave on
            unintentionally.

    Main GUI objects:
        qgrp (PyQt5.QtWidgets.QGroupBox)
    """

    def __init__(
        self,
        dev: Compax3_servo,
        DAQ_interval_ms=250,
        critical_not_alive_count=1,
        DAQ_timer_type=QtCore.Qt.TimerType.CoarseTimer,
        debug=False,
        **kwargs,
    ):
        super().__init__(dev, **kwargs)  # Pass kwargs onto QtCore.QObject()
        self.dev: Compax3_servo  # Enforce type: removes `_NoDevice()`

        self.create_worker_DAQ(
            DAQ_trigger=DAQ_TRIGGER.INTERNAL_TIMER,
            DAQ_function=self._DAQ_function,
            DAQ_interval_ms=DAQ_interval_ms,
            DAQ_timer_type=DAQ_timer_type,
            critical_not_alive_count=critical_not_alive_count,
            debug=debug,
        )

        self.create_worker_jobs(debug=debug)

        self._create_GUI()
        self._connect_signals_to_slots()
        self.signal_DAQ_updated.connect(self._update_GUI)

        if not self.dev.is_alive:
            self._update_GUI()  # Correctly reflect an offline device

        # Flags for Jog+/Jog- pushbutton control
        self._jog_plus_is_active = False
        self._jog_minus_is_active = False

    # --------------------------------------------------------------------------
    #   _DAQ_function
    # --------------------------------------------------------------------------

    def _DAQ_function(self) -> bool:
        success = self.dev.query_position()
        success &= self.dev.query_status_word_1()

        if not self.dev.status_word_1.no_error:
            self.dev.query_error()

        return success

    # --------------------------------------------------------------------------
    #   _create_GUI
    # --------------------------------------------------------------------------

    def _create_GUI(self):
        default_font_height = 17
        default_font_width = 8

        # Sub-groupbox: Status word 1 bits
        self.sw1_powerless = controls.create_tiny_LED()
        self.sw1_powered_stationary = controls.create_tiny_LED()
        self.sw1_zero_pos_known = controls.create_tiny_LED()
        self.sw1_pos_reached = controls.create_tiny_LED()

        i = 0
        p = {"parent": None, "alignment": QtCore.Qt.AlignmentFlag.AlignRight}
        grid = QtWid.QGridLayout()
        grid.setVerticalSpacing(4)
        # fmt: off
        grid.addWidget(QtWid.QLabel("powerless", **p)         , i, 0)
        grid.addWidget(self.sw1_powerless                     , i, 1); i+=1
        grid.addWidget(QtWid.QLabel("powered stationary", **p), i, 0)
        grid.addWidget(self.sw1_powered_stationary            , i, 1); i+=1
        grid.addWidget(QtWid.QLabel("zero pos. known", **p)   , i, 0)
        grid.addWidget(self.sw1_zero_pos_known                , i, 1); i+=1
        grid.addWidget(QtWid.QLabel("position reached", **p)  , i, 0)
        grid.addWidget(self.sw1_pos_reached                   , i, 1); i+=1
        # fmt: on
        # grid.setColumnStretch(0, 0)
        # grid.setColumnStretch(1, 0)

        self.qgrp_sw1 = QtWid.QGroupBox("Status word 1")
        self.qgrp_sw1.setLayout(grid)

        # Main groupbox
        font_lbl_status = QtGui.QFont(
            "Palatino", 14, weight=QtGui.QFont.Weight.Bold
        )
        self.lbl_status = QtWid.QLabel("OFFLINE")
        self.lbl_status.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        self.lbl_status.setFont(font_lbl_status)
        self.lbl_status.setFixedHeight(
            3 * QtGui.QFontMetrics(font_lbl_status).height()
        )

        self.sw1_error_tripped = controls.create_error_LED(text="No error")
        self.error_msg = QtWid.QPlainTextEdit("")
        self.error_msg.setLineWrapMode(
            QtWid.QPlainTextEdit.LineWrapMode.WidgetWidth
        )
        self.error_msg.setStyleSheet(controls.SS_TEXTBOX_ERRORS)
        self.error_msg.setMinimumWidth(22 * default_font_width)
        self.error_msg.setFixedHeight(4 * default_font_height)
        self.pbtn_ackn_error = QtWid.QPushButton("Acknowledge error")
        p = {"parent": None, "alignment": QtCore.Qt.AlignmentFlag.AlignRight}
        p2 = {**p, "readOnly": True}
        self.qlin_cur_pos = QtWid.QLineEdit("nan", **p2)
        self.qlin_new_pos = QtWid.QLineEdit("nan", **p)
        self.pbtn_move_to_new_pos = QtWid.QPushButton("Move to new position")
        self.pbtn_move_to_new_pos.setFixedHeight(3 * default_font_height)
        self.pbtn_jog_plus = QtWid.QPushButton("Jog +")
        self.pbtn_jog_minus = QtWid.QPushButton("Jog -")
        self.pbtn_stop = QtWid.QPushButton("\nSTOP &&\nREMOVE POWER\n")
        self.lbl_update_counter = QtWid.QLabel("0")

        i = 0
        grid = QtWid.QGridLayout()
        grid.setVerticalSpacing(4)
        # fmt: off
        grid.addWidget(self.lbl_status             , i, 0, 1, 3); i+=1
        grid.addItem(QtWid.QSpacerItem(1, 12)      , i, 0)      ; i+=1
        grid.addWidget(self.sw1_error_tripped      , i, 0, 1, 3); i+=1
        grid.addWidget(self.error_msg              , i, 0, 1, 3); i+=1
        grid.addWidget(self.pbtn_ackn_error        , i, 0, 1, 3); i+=1
        grid.addItem(QtWid.QSpacerItem(1, 12)      , i, 0)      ; i+=1
        grid.addWidget(self.qgrp_sw1               , i, 0, 1, 3); i+=1
        grid.addItem(QtWid.QSpacerItem(1, 12)      , i, 0)      ; i+=1

        grid.addWidget(QtWid.QLabel("Position:")   , i, 0, 1, 3); i+=1
        grid.addWidget(QtWid.QLabel("Current")     , i, 0)
        grid.addWidget(self.qlin_cur_pos           , i, 1)
        grid.addWidget(QtWid.QLabel("mm")          , i, 2)      ; i+=1
        grid.addWidget(QtWid.QLabel("New")         , i, 0)
        grid.addWidget(self.qlin_new_pos           , i, 1)
        grid.addWidget(QtWid.QLabel("mm")          , i, 2)      ; i+=1

        grid.addItem(QtWid.QSpacerItem(1, 12)      , i, 0)      ; i+=1
        grid.addWidget(self.pbtn_move_to_new_pos   , i, 0, 1, 3); i+=1
        grid.addWidget(self.pbtn_jog_plus          , i, 0, 1, 3); i+=1
        grid.addWidget(self.pbtn_jog_minus         , i, 0, 1, 3); i+=1
        grid.addWidget(self.pbtn_stop              , i, 0, 1, 3); i+=1
        grid.addWidget(self.lbl_update_counter     , i, 0, 1, 3); i+=1
        # fmt: on
        # grid.setColumnStretch(0, 0)
        # grid.setColumnStretch(1, 0)
        # grid.setColumnStretch(2, 0)

        self.qgrp = QtWid.QGroupBox(f"{self.dev.name}")
        self.qgrp.setStyleSheet(controls.SS_GROUP)
        self.qgrp.setLayout(grid)
        self.qgrp.setMaximumWidth(200)  # Work-around, hard limit width

    # --------------------------------------------------------------------------
    #   _update_GUI
    # --------------------------------------------------------------------------

    @Slot()
    def _update_GUI(self):
        """NOTE: 'self.dev.mutex' is not being locked, because we are only
        reading 'state' for displaying purposes. We can do this because 'state'
        members are written and read atomicly.
        Not locking the mutex might speed up the program.
        """
        SW = self.dev.status_word_1  # Shorthand
        if self.dev.is_alive:
            # At startup
            if self.update_counter_DAQ == 1:
                self.qlin_new_pos.setText(f"{self.dev.state.cur_pos:.2f}")

            if SW.powerless:
                self.lbl_status.setText("powerless")
            else:
                if SW.powered_stationary:
                    self.lbl_status.setText("POWERED\nstationary")
                else:
                    self.lbl_status.setText("POWERED")

            self.sw1_error_tripped.setChecked(not SW.no_error)
            if SW.no_error:
                self.sw1_error_tripped.setText("No error")
                self.error_msg.setPlainText("")
                self.error_msg.setReadOnly(False)
                self.error_msg.setStyleSheet(controls.SS_TEXTBOX_ERRORS)
            else:
                self.sw1_error_tripped.setText("ERROR TRIPPED")
                self.error_msg.setPlainText(self.dev.state.error_msg)
                self.error_msg.setReadOnly(True)
                self.error_msg.setStyleSheet(controls.SS_TEXTBOX_ERRORS)
            self.sw1_powerless.setChecked(bool(SW.powerless))
            self.sw1_powered_stationary.setChecked(bool(SW.powered_stationary))
            self.sw1_zero_pos_known.setChecked(bool(SW.zero_pos_known))
            self.sw1_pos_reached.setChecked(bool(SW.pos_reached))
            self.qlin_cur_pos.setText(f"{self.dev.state.cur_pos:.2f}")

            self.lbl_update_counter.setText(f"{self.update_counter_DAQ}")
        else:
            self.qgrp.setEnabled(False)

    # --------------------------------------------------------------------------
    #   GUI functions
    # --------------------------------------------------------------------------

    @Slot()
    def process_pbtn_ackn_error(self):
        self.send(self.dev.acknowledge_error)

    @Slot()
    def process_editingFinished_qlin_new_pos(self):
        try:
            new_pos = float(self.qlin_new_pos.text())
        except (TypeError, ValueError):
            new_pos = 0.0
        except Exception as e:
            raise e
        self.qlin_new_pos.setText(f"{new_pos:.2f}")

    @Slot()
    def process_pbtn_move_to_new_pos(self):
        # Double check if the value in the QLineEdit is actually numeric
        try:
            new_pos = float(self.qlin_new_pos.text())
        except Exception as e:
            raise e
        self.send(self.dev.move_to_target_position, (new_pos, 2))

    @Slot()
    def process_pbtn_jog_plus_pressed(self):
        if not self._jog_plus_is_active:
            self._jog_plus_is_active = True
            self.send(self.dev.jog_plus)

    @Slot()
    def process_pbtn_jog_plus_released(self):
        self._jog_plus_is_active = False
        self.send(self.dev.stop_motion_but_keep_power)

    @Slot()
    def process_pbtn_jog_minus_pressed(self):
        if not self._jog_minus_is_active:
            self._jog_minus_is_active = True
            self.send(self.dev.jog_minus)

    @Slot()
    def process_pbtn_jog_minus_released(self):
        self._jog_minus_is_active = False
        self.send(self.dev.stop_motion_but_keep_power)

    @Slot()
    def process_pbtn_stop(self):
        self.send(self.dev.stop_motion_and_remove_power)

    # --------------------------------------------------------------------------
    #   _connect_signals_to_slots
    # --------------------------------------------------------------------------

    def _connect_signals_to_slots(self):
        # self.send_setpoint.editingFinished.connect(
        #        self.send_setpoint_from_textbox)

        self.pbtn_ackn_error.clicked.connect(self.process_pbtn_ackn_error)
        self.qlin_new_pos.editingFinished.connect(
            self.process_editingFinished_qlin_new_pos
        )
        self.pbtn_move_to_new_pos.clicked.connect(
            self.process_pbtn_move_to_new_pos
        )
        # self.pbtn_debug.clicked.connect(self.process_pbtn_debug)
        self.pbtn_jog_plus.pressed.connect(self.process_pbtn_jog_plus_pressed)
        self.pbtn_jog_plus.released.connect(self.process_pbtn_jog_plus_released)
        self.pbtn_jog_minus.pressed.connect(self.process_pbtn_jog_minus_pressed)
        self.pbtn_jog_minus.released.connect(
            self.process_pbtn_jog_minus_released
        )
        self.pbtn_stop.clicked.connect(self.process_pbtn_stop)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""RS232 function library for a Parker Compax3 servo controller. Only the ASCII
version is supported, not (yet) the binary version.

Communication errors will be handled as non-fatal. This means it will struggle
on with the script while reporting error messages to the command line output,
as opposed to terminating the program completely.

State variables that read numpy.nan indicate that they are uninitialized or that
the previous query resulted in a communication error.

When this module is directly run from the terminal a demo will be shown.
"""
__author__ = "Dennis van Gils"
__authoremail__ = "vangils.dennis@gmail.com"
__url__ = "https://github.com/Dennis-van-Gils/python-dvg-devices"
__date__ = "23-05-2024"
__version__ = "1.5.0"
# pylint: disable=missing-function-docstring, pointless-string-statement

import sys
from typing import Union, Tuple

import numpy as np

from dvg_devices.BaseDevice import SerialDevice


class Compax3_servo(SerialDevice):
    class Status_word_1:
        """[numpy.nan] values indicate that the parameter is not initialized or
        that the last query was unsuccessful in communication."""

        # fmt: off
        I0                : Union[float, bool] = np.nan  # bit 0
        I1                : Union[float, bool] = np.nan  # bit 1
        I2                : Union[float, bool] = np.nan  # bit 2
        I3                : Union[float, bool] = np.nan  # bit 3
        I4                : Union[float, bool] = np.nan  # bit 4
        I5                : Union[float, bool] = np.nan  # bit 5
        I6                : Union[float, bool] = np.nan  # bit 6
        I7                : Union[float, bool] = np.nan  # bit 7 'open motor holding brake'
        no_error          : Union[float, bool] = np.nan  # bit 8
        pos_reached       : Union[float, bool] = np.nan  # bit 9
        powerless         : Union[float, bool] = np.nan  # bit 10
        powered_stationary: Union[float, bool] = np.nan  # bit 11 'standstill'
        zero_pos_known    : Union[float, bool] = np.nan  # bit 12
        PSB0              : Union[float, bool] = np.nan  # bit 13
        PSB1              : Union[float, bool] = np.nan  # bit 14
        PSB2              : Union[float, bool] = np.nan  # bit 15
        # fmt: on

    class State:
        """[numpy.nan] values indicate that the parameter is not initialized or
        that the last query was unsuccessful in communication."""

        # fmt: off
        cur_pos: float = np.nan  # position [mm]
        error_msg: str = ""      # error string message
        # fmt: on

    def __init__(
        self,
        name="trav",
        long_name="Compax3 servo",
        connect_to_serial_number=None,
    ):
        super().__init__(name=name, long_name=long_name)

        # Default serial settings
        self.serial_settings = {
            "baudrate": 115200,
            "timeout": 0.4,
            "write_timeout": 0.4,
            "rtscts": True,
        }
        self.set_read_termination("\r")
        self.set_write_termination("\r")

        self.set_ID_validation_query(
            ID_validation_query=self.ID_validation_query,
            valid_ID_broad="Compax3",
            valid_ID_specific=connect_to_serial_number,
        )

        # Containers for the status, process and measurement variables
        self.status_word_1 = self.Status_word_1()
        self.state = self.State()

        self.serial_str = ""  # Serial number of the Compax3

    # --------------------------------------------------------------------------
    #   OVERRIDE: query
    # --------------------------------------------------------------------------

    def query(
        self,
        msg: Union[str, bytes],
        raises_on_timeout: bool = False,
        returns_ascii: bool = True,
    ) -> Tuple[bool, Union[str, bytes, None]]:
        success, reply = super().query(msg, raises_on_timeout, returns_ascii)

        # The Compax3 is more complex in its replies than the average device.
        # Hence:
        if isinstance(reply, str):
            if reply[0] == ">":
                # Successful operation without meaningful reply
                pass
            elif reply[0] == "!":
                # Error reply
                print("COMPAX3 COMMUNICATION ERROR: " + reply)
                success = False
            else:
                # Successful operation with meaningful reply
                pass

        return success, reply

    # --------------------------------------------------------------------------
    #   ID_validation_query
    # --------------------------------------------------------------------------

    def ID_validation_query(self) -> Tuple[str, Union[str, None]]:
        broad_reply = ""
        specific_reply = None

        _success, reply = self.query("_?")
        if isinstance(reply, str):
            broad_reply = reply[:7]  # Expected: "Compax3"

        _success, reply = self.query("o1.4")
        if isinstance(reply, str):
            specific_reply = reply  # Serial number

        return broad_reply, specific_reply

    # --------------------------------------------------------------------------
    #   begin
    # --------------------------------------------------------------------------

    def begin(self):
        """This function should run directly after having established a
        connection to the device.

        Returns: True if successful, False otherwise.
        """
        success = self.query_serial_str()
        success &= self.query_error()
        success &= self.query_position()
        success &= self.query_status_word_1()
        return success

    # --------------------------------------------------------------------------
    #   Higher level queries
    # --------------------------------------------------------------------------

    def query_serial_str(self) -> bool:
        """Query the serial number and store it in the class member 'serial_str'

        Returns: True if successful, False otherwise.
        """
        # First make sure we're dealing with a Compax3 controller, because we
        # have to exclude the possibility that another device will reply with
        # an error message to the serial number request, which then could be
        # mistaken for /the/ serial number.
        _success, reply = self.query("_?")
        if isinstance(reply, str) and reply.startswith("Compax3"):
            # Now we can query the serial number
            _success, reply = self.query("o1.4")
            if isinstance(reply, str):
                self.serial_str = reply
                return True

        self.serial_str = ""
        return False

    def query_position(self) -> bool:
        """Query the position and store in the class member 'state.cur_pos'
        when successful. When the communication fails the class member will be
        set to [numpy.nan].

        Returns: True if successful, False otherwise.
        """
        _success, reply = self.query("o680.5")
        if isinstance(reply, str):
            self.state.cur_pos = float(reply)
            return True

        self.state.cur_pos = np.nan
        return False

    def query_error(self) -> bool:
        """Query the last error and store in the class member 'state.error_msg'
        when successful. When the communication fails the class member will be
        set to [numpy.nan].

        Returns: True if successful, False otherwise.
        """
        _success, reply = self.query("o550.1")
        if isinstance(reply, str):
            # Translate error codes to more meaningful messages
            if reply == "1":
                self.state.error_msg = ""
            elif reply == "17168":
                self.state.error_msg = f"{reply}: Motor temperature"
            elif reply == "29472":
                self.state.error_msg = f"{reply}: Following error"
            elif reply == "29475":
                self.state.error_msg = (
                    f"{reply}: Target or actual position "
                    "exceeds positive end limit"
                )
            elif reply == "29476":
                self.state.error_msg = (
                    f"{reply}: Target or actual position "
                    "exceeds negative end limit"
                )
            elif reply == "29479":
                self.state.error_msg = (
                    f"{reply}: Change of direction during " "movement"
                )
            else:
                self.state.error_msg = reply
            return True

        self.state.error_msg = ""
        return False

    def query_status_word_1(self) -> bool:
        """Query the status word 1 and store in the class member 'status_word_1'
        when successful. When the communication fails the class member
        'status_word_1' will be populated with [numpy.nan].

        Returns: True if successful, False otherwise.
        """
        success, reply = self.query("o1000.3")

        if reply is None:
            # fmt: off
            self.status_word_1.I0 = np.nan
            self.status_word_1.I1 = np.nan
            self.status_word_1.I2 = np.nan
            self.status_word_1.I3 = np.nan
            self.status_word_1.I4 = np.nan
            self.status_word_1.I5 = np.nan
            self.status_word_1.I6 = np.nan
            self.status_word_1.I7 = np.nan
            self.status_word_1.no_error           = np.nan
            self.status_word_1.pos_reached        = np.nan
            self.status_word_1.powerless          = np.nan
            self.status_word_1.powered_stationary = np.nan
            self.status_word_1.zero_pos_known     = np.nan
            self.status_word_1.PSB0 = np.nan
            self.status_word_1.PSB1 = np.nan
            self.status_word_1.PSB2 = np.nan
            # fmt: on
        else:
            dec_x = int(reply)

            # Convert dec to bin string, remove prefix '0b', prefix with 0's to
            # garantuee 16 bits and reverse
            str_bits = ((bin(dec_x)[2:]).zfill(16))[::-1]

            # fmt: off
            self.status_word_1.I0 = bool(int(str_bits[0]))
            self.status_word_1.I1 = bool(int(str_bits[1]))
            self.status_word_1.I2 = bool(int(str_bits[2]))
            self.status_word_1.I3 = bool(int(str_bits[3]))
            self.status_word_1.I4 = bool(int(str_bits[4]))
            self.status_word_1.I5 = bool(int(str_bits[5]))
            self.status_word_1.I6 = bool(int(str_bits[6]))
            self.status_word_1.I7 = bool(int(str_bits[7]))
            self.status_word_1.no_error           = bool(int(str_bits[8]))
            self.status_word_1.pos_reached        = bool(int(str_bits[9]))
            self.status_word_1.powerless          = bool(int(str_bits[10]))
            self.status_word_1.powered_stationary = bool(int(str_bits[11]))
            self.status_word_1.zero_pos_known     = bool(int(str_bits[12]))
            self.status_word_1.PSB0 = bool(int(str_bits[13]))
            self.status_word_1.PSB1 = bool(int(str_bits[14]))
            self.status_word_1.PSB2 = bool(int(str_bits[15]))
            # fmt: on

        return success

    # --------------------------------------------------------------------------
    # --------------------------------------------------------------------------

    def store_motion_profile(
        self,
        target_position=0,
        velocity=10,
        mode=1,
        accel=100,
        decel=100,
        jerk=1e6,
        profile_number=2,
    ) -> bool:
        """
        Note:
            Profile_number 0 is reserved for homing.
            Movement mode is fixed to absolute, not relative.
        """
        mode = 1  # Overrule, set movement mode to absolute position
        p = profile_number  # Shorthand

        print(f"  Profile number: {p}")
        print(f"    pos   = {target_position:.2f}")
        print(f"    vel   = {velocity:.2f}")
        print(f"    mode  = {mode:d} (fixed to absolute)")
        print(f"    accel = {accel:.2f}")
        print(f"    decel = {decel:.2f}")
        print(f"    jerk  = {jerk:.2f}\n")

        success, _reply = self.query(f"o1901.{p:d}={target_position:.2f}")
        if success:
            success, _reply = self.query(f"o1902.{p:d}={velocity:.2f}")
        if success:
            success, _reply = self.query(f"o1905.{p:d}={mode:d}")
        if success:
            success, _reply = self.query(f"o1906.{p:d}={accel:.2f}")
        if success:
            success, _reply = self.query(f"o1907.{p:d}={decel:.2f}")
        if success:
            success, _reply = self.query(f"o1908.{p:d}={jerk:.2f}")
        if success:
            success, _reply = self.query(f"o1904.{p:d}=$32")  # Store profile

        return success

    def activate_motion_profile(self, profile_number=2) -> bool:
        # Control word (CW) for activating the passed profile number
        # First send: quit/motor bit (bit 0) high
        #             stop bits (bits 1, 14) high
        #             start bit (bit 13) low
        CW_LO = 0b0100000000000011
        CW_LO = CW_LO + (profile_number << 8)
        success, _reply = self.query(f"o1100.3={CW_LO:d}")
        if success:
            # Then send start bit (bit 13) high
            CW_HI = CW_LO + (1 << 13)
            success, _reply = self.query(f"o1100.3={CW_HI:d}")

        return success

    def move_to_target_position(
        self, target_position, profile_number=2
    ) -> bool:
        """
        Note: Make sure a motion profile with number 'profile_number' is stored
        at least once with 'self.store_motion_profile' before moving.
        """
        # Send new target position
        success, _reply = self.query(
            f"o1901.{profile_number:d}={target_position:.2f}"
        )

        if success:
            self.activate_motion_profile(profile_number=2)

        return success

    def jog_plus(self) -> bool:
        # Control word (CW) for activating the jog+
        CW_LO = 0b0100000000000011
        success, _reply = self.query(f"o1100.3={CW_LO:d}")
        if success:
            # Then send jog+ bit (bit 2) high
            CW_HI = CW_LO + (1 << 2)
            success, _reply = self.query(f"o1100.3={CW_HI:d}")

        return success

    def jog_minus(self) -> bool:
        # Control word (CW) for activating the jog-
        CW_LO = 0b0100000000000011
        success, _reply = self.query(f"o1100.3={CW_LO:d}")
        if success:
            # Then send jog- bit (bit 3) high
            CW_HI = CW_LO + (1 << 3)
            success, _reply = self.query(f"o1100.3={CW_HI:d}")

        return success

    def stop_motion_but_keep_power(self) -> bool:
        CW_LO = 0b0100000000000011
        success, _reply = self.query(f"o1100.3={CW_LO:d}")

        return success

    def stop_motion_and_remove_power(self) -> bool:
        success, _reply = self.query("o1100.3=0")

        return success

    def acknowledge_error(self) -> bool:
        """If the cause of an error is eliminated, the error can be
        acknowledged. This is necessary for the axis to be able to get powered
        again.

        Sends a rising edge on bit 0 of the control word. This will leave the
        axis powered when no new error occurs again.

        Returns: True if successful, False otherwise.
        """
        success, _reply = self.query("o1100.3=0")
        if success:
            success, _reply = self.query("o1100.3=1")

        return success

    def report_status_word_1(self, compact=False):
        SW = self.status_word_1  # Shorthand
        if not compact:
            print("Status word 1:")
            print(f"  {SW.I0:-6s}: I0")
            print(f"  {SW.I1:-6s}: I1")
            print(f"  {SW.I2:-6s}: I2")
            print(f"  {SW.I3:-6s}: I3")
            print(f"  {SW.I4:-6s}: I4")
            print(f"  {SW.I5:-6s}: I5")
            print(f"  {SW.I6:-6s}: I6")
            print(f"  {SW.I7:-6s}: I7")
            print(f"  {SW.no_error:-6s}: no_error")
            print(f"  {SW.pos_reached:-6s}: pos_reached")
            print(f"  {SW.powerless:-6s}: powerless")
            print(f"  {SW.powered_stationary:-6s}: powered_stat")
            print(f"  {SW.zero_pos_known:-6s}: zero_pos_known")
            print(f"  {SW.PSB0:-6s}: PSB0")
            print(f"  {SW.PSB1:-6s}: PSB1")
            print(f"  {SW.PSB2:-6s}: PSB2")
        else:
            if not SW.no_error:
                print("  ERROR!")
            if SW.powerless:
                print("  Axis    : unpowered")
            else:
                if SW.powered_stationary:
                    print("  Axis    : POWERED STANDSTILL")
                else:
                    print("  Axis    : POWERED")
            if SW.zero_pos_known:
                print("  Zero pos: known")
            else:
                print("  Zero pos: UNKNOWN")
            if SW.pos_reached:
                print("  Position: REACHED")
            else:
                print("  Position: unreached")
        print("")


# ------------------------------------------------------------------------------
#   Main: Will show a demo when run from the terminal
# ------------------------------------------------------------------------------

if __name__ == "__main__":
    import time

    # Specific connection settings of each traverse axis of our setup
    class Trav_connection_params:
        # Serial number of the Compax3 servo controller to connect to.
        # Set to "" to connect to any Compax3.
        serial = ""
        # Display name
        name = "TRAV"
        # Path to the config textfile containing the (last used) RS232 port
        path_config = "config/port_Compax3_servo.txt"

    # Horizontal axis
    trav_conn_horz = Trav_connection_params()
    trav_conn_horz.serial = "4409980001"
    trav_conn_horz.name = "TRAV HORZ"
    trav_conn_horz.path_config = "config/port_Compax3_servo_horz.txt"

    # Vertical axis
    trav_conn_vert = Trav_connection_params()
    trav_conn_vert.serial = "4319370001"
    trav_conn_vert.name = "TRAV VERT"
    trav_conn_vert.path_config = "config/port_Compax3_servo_vert.txt"

    # Connect to this specific traverse axis
    # trav_conn = Trav_connection_params()  # Any
    # trav_conn = trav_conn_horz
    trav_conn = trav_conn_vert

    # --------------------------------------------------------------------------
    # --------------------------------------------------------------------------

    # Create connection to Compax3 servo controller over RS232
    trav = Compax3_servo(
        name=trav_conn.name,
        connect_to_serial_number=trav_conn.serial,
    )

    if trav.auto_connect(filepath_last_known_port=trav_conn.path_config):
        trav.begin()  # Retrieve necessary parameters
    else:
        time.sleep(1)
        sys.exit(0)

    print(f"Error msg: {trav.state.error_msg}")
    print(f"Current position: {trav.state.cur_pos:.2f}")
    trav.report_status_word_1(compact=True)

    trav.acknowledge_error()

    print("ERRORS RESET")
    print("ACTIVATE\n")
    time.sleep(0.1)

    trav.query_status_word_1()
    trav.report_status_word_1(compact=True)

    # Update set #2
    # fmt: off
    trav.query("o1901.2=-180.0")     # target position
    trav.query("o1902.2=100.0")      # velocity
    trav.query("o1905.2=1")          # mode: 1 (MoveAbs), 2 (MoveRel)
    trav.query("o1906.2=100")        # accel
    trav.query("o1907.2=100")        # decel
    trav.query("o1908.2=1000000")    # jerk
    trav.query("o1904.2=$32")        # store profile

    trav.query("o1100.3=16899")      # set #2, start bit low
    trav.query("o1100.3=25091")      # set #2, start bit high

    """
    trav.query("o1100.3=$4003")      # set #0, homing, start bit low
    trav.query("o1100.3=$6003")      # set #0, homing, start bit high
    """

    """
    trav.query("o1100.3=$4007")      # jog+
    time.sleep(2);

    trav.query_position()
    print(f"Current position: {trav.state.cur_pos:.2f}")

    trav.query("o1100.3=$400b")      # jog-
    time.sleep(2);
    """
    # fmt: on

    print("MOVING\n")

    for i in range(14):
        trav.query_position()
        print(f"Current position: {trav.state.cur_pos:.2f}")
        trav.query_status_word_1()
        trav.report_status_word_1(compact=True)
        time.sleep(0.2)

    trav.query("o1100.3=0")  # disable axis
    # trav.query("o1000.4")   # last executed set #

    print("DEACTIVATE\n")
    time.sleep(1)

    trav.query_error()
    trav.query_position()
    trav.query_status_word_1()

    print(f"Error msg: {trav.state.error_msg}")
    print(f"Current position: {trav.state.cur_pos:.2f}")
    trav.report_status_word_1(compact=True)

    # Close
    print("")
    trav.close()
    sys.exit(0)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""PyQt/Pyside module to provide multithreaded communication and periodical data
acquisition for a Xylem Hydrovar HVL variable speed pump controller.
"""
__author__ = "Dennis van Gils"
__authoremail__ = "vangils.dennis@gmail.com"
__url__ = "https://github.com/Dennis-van-Gils/python-dvg-devices"
__date__ = "23-05-2024"
__version__ = "1.5.0"
# pylint: disable=broad-except, missing-function-docstring, multiple-statements

import time
from enum import IntEnum

from qtpy import QtCore, QtWidgets as QtWid
from qtpy.QtCore import Signal, Slot  # type: ignore

from dvg_debug_functions import dprint, print_fancy_traceback as pft
import dvg_pyqt_controls as controls
from dvg_qdeviceio import QDeviceIO, DAQ_TRIGGER
from dvg_devices.XylemHydrovarHVL_protocol_RTU import XylemHydrovarHVL, HVL_Mode


# Enumeration
class GUI_input_fields(IntEnum):
    ALL = 0
    P_WANTED = 1
    F_WANTED = 2
    HVL_MODE = 3


class XylemHydrovarHVL_qdev(QDeviceIO):
    """Manages multithreaded communication and periodical data acquisition for
    a Xylem Hydrovar HVL variable speed drive (VSD) controller.

    In addition, it also provides PyQt/PySide GUI objects for control of the
    device. These can be incorporated into your application.

    All device I/O operations will be offloaded to 'workers', each running in
    a newly created thread.

    (*): See 'dvg_qdeviceio.QDeviceIO()' for details.

    Args:
        dev:
            Reference to a 'XylemHydrovarHVL_protocol_RTU.XylemHydrovarHVL'
            instance.

        debug:
            Show debug info in terminal? Warning: Slow! Do not leave on
            unintentionally.

    Main GUI objects:
        qgrp_control (PyQt5.QtWidgets.QGroupBox)
        qgrp_inverter (PyQt5.QtWidgets.QGroupBox)
        qgrp_error_status (PyQt5.QtWidgets.QGroupBox)
    """

    signal_GUI_input_field_update = Signal(int)
    signal_pump_just_stopped_and_reached_standstill = Signal()

    def __init__(
        self,
        dev: XylemHydrovarHVL,
        DAQ_trigger=DAQ_TRIGGER.INTERNAL_TIMER,
        DAQ_interval_ms=200,
        DAQ_timer_type=QtCore.Qt.TimerType.PreciseTimer,
        critical_not_alive_count=3,
        debug=False,
        **kwargs,
    ):
        super().__init__(dev, **kwargs)  # Pass kwargs onto QtCore.QObject()
        self.dev: XylemHydrovarHVL  # Enforce type: removes `_NoDevice()`

        self.create_worker_DAQ(
            DAQ_trigger=DAQ_trigger,
            DAQ_function=self._DAQ_function,
            DAQ_interval_ms=DAQ_interval_ms,
            DAQ_timer_type=DAQ_timer_type,
            critical_not_alive_count=critical_not_alive_count,
            debug=debug,
        )
        self.create_worker_jobs(jobs_function=self._jobs_function, debug=debug)

        self._create_GUI()
        self.signal_DAQ_updated.connect(self._update_GUI)
        self.signal_connection_lost.connect(self._update_GUI)
        self.signal_GUI_input_field_update.connect(self._update_GUI_input_field)

        self._update_GUI()
        self._update_GUI_input_field()

        # Mechanism to detect that the pump has received a stop command and has
        # reached full standstill.
        # ----------------------------------------------------------------------
        # Below flag is True as soon as the `pump stop` command is send. It will
        # reset to False as soon as full standstill is achieved, after which
        # signal `signal_pump_just_stopped_and_reached_standstill` will be
        # emitted.
        self.pump_is_stopping = False

    # --------------------------------------------------------------------------
    #   _DAQ_function
    # --------------------------------------------------------------------------

    def _DAQ_function(self) -> bool:
        """Every DAQ time step, read the following:
        - error status
        - device status
        - actual pressure
        - actual inverter frequency

        Every N'th DAQ time step, read the following:
        - inverter diagnostics
        """
        DEBUG_local = False
        if DEBUG_local:
            tick = time.perf_counter()

        success = self.dev.read_error_status()
        success &= self.dev.read_device_status()
        success &= self.dev.read_actual_pressure()
        success &= self.dev.read_actual_frequency()

        if (self.update_counter_DAQ % 5) == 0:
            success &= self.dev.read_inverter_diagnostics()

        if self.pump_is_stopping:
            if self.dev.state.pump_is_running:
                # Pump is still coasting down
                pass
            else:
                # Reached standstill
                self.pump_is_stopping = False
                self.signal_pump_just_stopped_and_reached_standstill.emit()
                print("Pump reached standstill")

        if not success:
            return False

        if DEBUG_local:
            tock = time.perf_counter()
            dprint(f"{self.dev.name}: done in {tock - tick:.3f}")

        return True

    # --------------------------------------------------------------------------
    #   _jobs_function
    # --------------------------------------------------------------------------

    def _jobs_function(self, func, args):
        if func == "signal_GUI_input_field_update":
            # Special instruction
            self.signal_GUI_input_field_update.emit(*args)
        else:
            # Default job processing:
            # Send I/O operation to the device
            try:
                func(*args)
            except Exception as err:
                pft(err)

    # --------------------------------------------------------------------------
    #   _create_GUI
    # --------------------------------------------------------------------------

    def _create_GUI(self):
        p = {
            "alignment": QtCore.Qt.AlignmentFlag.AlignRight,
            "minimumWidth": controls.e8(8),
            "maximumWidth": controls.e8(8),
        }
        p2 = {**p, "readOnly": True}

        # Pump control
        self.pbtn_pump_onoff = controls.create_Toggle_button("OFFLINE")
        self.indicator_pump_running = controls.create_tiny_LED()
        self.rbtn_mode_pressure = QtWid.QRadioButton("Regulate pressure")
        self.rbtn_mode_frequency = QtWid.QRadioButton("Fixed frequency")
        self.qled_P_wanted = QtWid.QLineEdit("nan", **p)
        self.qled_P_actual = QtWid.QLineEdit("nan", **p2)
        self.qled_P_limits = QtWid.QLineEdit(
            f"0 \u2013 {self.dev.max_pressure_setpoint_bar:.2f}", **p2
        )
        self.qled_f_wanted = QtWid.QLineEdit("nan", **p)
        self.qled_f_actual = QtWid.QLineEdit("nan", **p2)
        self.qled_f_limits = QtWid.QLineEdit(
            f"{self.dev.state.min_frequency:.0f} \u2013 "
            f"{self.dev.state.max_frequency:.0f}",
            **p2,
        )
        self.qlbl_update_counter = QtWid.QLabel("0")

        self.pbtn_pump_onoff.clicked.connect(self._process_pbtn_pump_onoff)
        self.rbtn_mode_pressure.clicked.connect(self._process_rbtn_mode)
        self.rbtn_mode_frequency.clicked.connect(self._process_rbtn_mode)
        self.qled_P_wanted.editingFinished.connect(
            self._send_P_wanted_from_textbox
        )
        self.qled_f_wanted.editingFinished.connect(
            self._send_f_wanted_from_textbox
        )

        # fmt: off
        i = 0
        grid = QtWid.QGridLayout()
        grid.setVerticalSpacing(4)

        grid.addWidget(self.pbtn_pump_onoff            , i, 0, 1, 3); i+=1
        grid.addWidget(QtWid.QLabel("Pump running?")   , i, 0, 1, 2)
        grid.addWidget(self.indicator_pump_running     , i, 2)      ; i+=1
        grid.addItem(QtWid.QSpacerItem(1, 10)          , i, 0)      ; i+=1

        grid.addWidget(QtWid.QLabel("<b>Mode</b>")     , i, 0, 1, 3); i+=1
        grid.addItem(QtWid.QSpacerItem(1, 6)           , i, 0)      ; i+=1
        grid.addWidget(self.rbtn_mode_pressure         , i, 0, 1, 3); i+=1
        grid.addWidget(self.rbtn_mode_frequency        , i, 0, 1, 3); i+=1
        grid.addItem(QtWid.QSpacerItem(1, 10)          , i, 0)      ; i+=1

        grid.addWidget(QtWid.QLabel("<b>Pressure</b>") , i, 0, 1, 3); i+=1
        grid.addItem(QtWid.QSpacerItem(1, 6)           , i, 0)      ; i+=1
        grid.addWidget(QtWid.QLabel("Wanted")          , i, 0)
        grid.addWidget(self.qled_P_wanted              , i, 1)
        grid.addWidget(QtWid.QLabel("bar")             , i, 2)      ; i+=1
        grid.addWidget(QtWid.QLabel("Actual")          , i, 0)
        grid.addWidget(self.qled_P_actual              , i, 1)
        grid.addWidget(QtWid.QLabel("bar")             , i, 2)      ; i+=1
        grid.addWidget(QtWid.QLabel("Limits")          , i, 0)
        grid.addWidget(self.qled_P_limits              , i, 1)
        grid.addWidget(QtWid.QLabel("bar")             , i, 2)      ; i+=1
        grid.addItem(QtWid.QSpacerItem(1, 10)          , i, 0)      ; i+=1

        grid.addWidget(QtWid.QLabel("<b>Frequency</b>"), i, 0, 1, 3); i+=1
        grid.addItem(QtWid.QSpacerItem(1, 6)           , i, 0)      ; i+=1
        grid.addWidget(QtWid.QLabel("Wanted")          , i, 0)
        grid.addWidget(self.qled_f_wanted              , i, 1)
        grid.addWidget(QtWid.QLabel("Hz")              , i, 2)      ; i+=1
        grid.addWidget(QtWid.QLabel("Actual")          , i, 0)
        grid.addWidget(self.qled_f_actual              , i, 1)
        grid.addWidget(QtWid.QLabel("Hz")              , i, 2)      ; i+=1
        grid.addWidget(QtWid.QLabel("Limits")          , i, 0)
        grid.addWidget(self.qled_f_limits              , i, 1)
        grid.addWidget(QtWid.QLabel("Hz")              , i, 2)      ; i+=1
        grid.addItem(QtWid.QSpacerItem(1, 10)          , i, 0)      ; i+=1
        grid.addWidget(self.qlbl_update_counter        , i, 0)
        # fmt: on

        grid.setColumnStretch(0, 0)
        grid.setColumnStretch(1, 0)
        grid.setColumnStretch(2, 1)
        grid.setAlignment(QtCore.Qt.AlignmentFlag.AlignTop)

        self.qgrp_control = QtWid.QGroupBox("Pump control")
        self.qgrp_control.setLayout(grid)

        # Inverter diagnostics
        self.qled_inverter_temp = QtWid.QLineEdit("nan", **p2)
        self.qled_inverter_curr_A = QtWid.QLineEdit("nan", **p2)
        self.qled_inverter_curr_pct = QtWid.QLineEdit("nan", **p2)
        self.qled_inverter_volt = QtWid.QLineEdit("nan", **p2)

        # fmt: off
        i = 0
        grid = QtWid.QGridLayout()
        grid.setVerticalSpacing(4)

        grid.addWidget(QtWid.QLabel("Temp.")           , i, 0)
        grid.addWidget(self.qled_inverter_temp         , i, 1)
        grid.addWidget(QtWid.QLabel("\u00b0C")         , i, 2)      ; i+=1
        grid.addWidget(QtWid.QLabel("Voltage")         , i, 0)
        grid.addWidget(self.qled_inverter_volt         , i, 1)
        grid.addWidget(QtWid.QLabel("V")               , i, 2)      ; i+=1
        grid.addWidget(QtWid.QLabel("Current")         , i, 0)
        grid.addWidget(self.qled_inverter_curr_A       , i, 1)
        grid.addWidget(QtWid.QLabel("A")               , i, 2)      ; i+=1
        grid.addWidget(QtWid.QLabel("Current")         , i, 0)
        grid.addWidget(self.qled_inverter_curr_pct     , i, 1)
        grid.addWidget(QtWid.QLabel("%")               , i, 2)      ; i+=1
        # fmt: on

        grid.setColumnStretch(0, 0)
        grid.setColumnStretch(1, 0)
        grid.setColumnStretch(2, 1)
        grid.setAlignment(QtCore.Qt.AlignmentFlag.AlignTop)

        self.qgrp_inverter = QtWid.QGroupBox("Inverter")
        self.qgrp_inverter.setLayout(grid)

        # Error status
        self.qpte_error_status = QtWid.QPlainTextEdit()
        self.qpte_error_status.setStyleSheet(controls.SS_TEXTBOX_ERRORS)

        grid = QtWid.QGridLayout()
        grid.setVerticalSpacing(4)
        grid.addWidget(self.qpte_error_status, 0, 0)
        grid.setAlignment(QtCore.Qt.AlignmentFlag.AlignTop)

        self.qgrp_error_status = QtWid.QGroupBox("Error status")
        self.qgrp_error_status.setLayout(grid)

    # --------------------------------------------------------------------------
    #   _update_GUI
    # --------------------------------------------------------------------------

    @Slot()
    def _update_GUI(self):
        """NOTE: 'self.dev.mutex' is not being locked, because we are only
        reading 'state' for displaying purposes. We can do this because 'state'
        members are written and read atomicly.
        Not locking the mutex might speed up the program.
        """

        if not self.dev.is_alive:
            self.qgrp_control.setEnabled(False)
            self.qgrp_inverter.setEnabled(False)
            self.qgrp_error_status.setEnabled(False)
            self.pbtn_pump_onoff.setText("OFFLINE")
            return

        # Shorthand
        state = self.dev.state
        error_status = self.dev.error_status

        # Pump control
        if state.pump_is_enabled:
            self.pbtn_pump_onoff.setEnabled(True)
            if state.pump_is_on:
                self.pbtn_pump_onoff.setChecked(True)
                self.pbtn_pump_onoff.setText("Pump is ON")
            else:
                self.pbtn_pump_onoff.setChecked(False)
                self.pbtn_pump_onoff.setText("Pump is OFF")
        else:
            self.pbtn_pump_onoff.setEnabled(False)
            self.pbtn_pump_onoff.setChecked(False)
            self.pbtn_pump_onoff.setText("Pump is DISABLED")

        self.indicator_pump_running.setChecked(state.pump_is_running)
        self.qled_P_actual.setText(f"{state.actual_pressure:.2f}")
        self.qled_f_actual.setText(f"{state.actual_frequency:.1f}")

        # Inverter diagnostics
        self.qled_inverter_temp.setText(f"{state.inverter_temp:.0f}")
        self.qled_inverter_volt.setText(f"{state.inverter_volt:.0f}")
        self.qled_inverter_curr_A.setText(f"{state.inverter_curr_A:.2f}")
        self.qled_inverter_curr_pct.setText(f"{state.inverter_curr_pct:.0f}")

        # Error status
        if error_status.has_error():
            error_text = ""
            if error_status.overcurrent:
                error_text += "#11: OVERCURRENT\n"
            if error_status.overload:
                error_text += "#12: OVERLOAD"
            if error_status.overvoltage:
                error_text += "#13: OVERVOLTAGE"
            if error_status.phase_loss:
                error_text += "#16: PHASE LOSS"
            if error_status.inverter_overheat:
                error_text += "#14: INVERTER OVERHEAT"
            if error_status.motor_overheat:
                error_text += "#15: MOTOR OVERHEAT"
            if error_status.lack_of_water:
                error_text += "#21: LACK OF WATER"
            if error_status.minimum_threshold:
                error_text += "#22: MINIMUM THRESHOLD"
            if error_status.act_val_sensor_1:
                error_text += "#23: ACT VAL SENSOR 1"
            if error_status.act_val_sensor_2:
                error_text += "#24: ACT VAL SENSOR 2"
            if error_status.setpoint_1_low_mA:
                error_text += "#25: SETPOINT 1 I<4 mA"
            if error_status.setpoint_2_low_mA:
                error_text += "#26: SETPOINT 2 I<4 mA"
            self.qpte_error_status.setReadOnly(True)
            self.qpte_error_status.setPlainText(error_text)
        else:
            if self.dev.device_status.device_has_a_warning:
                self.qpte_error_status.setReadOnly(True)
                self.qpte_error_status.setPlainText("WARNING!!!")
            else:
                self.qpte_error_status.setReadOnly(False)
                self.qpte_error_status.setPlainText("No errors")

        # Update counter
        self.qlbl_update_counter.setText(f"{self.update_counter_DAQ}")

    # --------------------------------------------------------------------------
    #   _update_GUI_input_field
    # --------------------------------------------------------------------------

    @Slot()
    @Slot(int)
    def _update_GUI_input_field(self, GUI_input_field=GUI_input_fields.ALL):
        if GUI_input_field == GUI_input_fields.P_WANTED:
            self.qled_P_wanted.setText(f"{self.dev.state.wanted_pressure:.2f}")

        elif GUI_input_field == GUI_input_fields.F_WANTED:
            self.qled_f_wanted.setText(f"{self.dev.state.wanted_frequency:.1f}")

        elif GUI_input_field == GUI_input_fields.HVL_MODE:
            if self.dev.state.hvl_mode == HVL_Mode.CONTROLLER:
                self.rbtn_mode_pressure.setChecked(True)
            if self.dev.state.hvl_mode == HVL_Mode.ACTUATOR:
                self.rbtn_mode_frequency.setChecked(True)

        else:
            self.qled_P_wanted.setText(f"{self.dev.state.wanted_pressure:.2f}")
            self.qled_f_wanted.setText(f"{self.dev.state.wanted_frequency:.1f}")
            if self.dev.state.hvl_mode == HVL_Mode.CONTROLLER:
                self.rbtn_mode_pressure.setChecked(True)
            if self.dev.state.hvl_mode == HVL_Mode.ACTUATOR:
                self.rbtn_mode_frequency.setChecked(True)

    # --------------------------------------------------------------------------
    #   GUI functions
    # --------------------------------------------------------------------------

    @Slot()
    def _process_pbtn_pump_onoff(self):
        if self.dev.state.pump_is_on:
            self.send_pump_stop()
        else:
            self.send_pump_start()

    @Slot()
    def _send_P_wanted_from_textbox(self):
        try:
            P_bar = float(self.qled_P_wanted.text())
        except (TypeError, ValueError):
            P_bar = 0.0
        except Exception as err:
            raise err

        self.add_to_jobs_queue(self.dev.set_wanted_pressure, P_bar)
        self.add_to_jobs_queue(
            "signal_GUI_input_field_update", GUI_input_fields.P_WANTED
        )
        self.process_jobs_queue()

    @Slot()
    def _send_f_wanted_from_textbox(self):
        try:
            f_Hz = float(self.qled_f_wanted.text())
        except (TypeError, ValueError):
            f_Hz = self.dev.state.min_frequency
        except Exception as err:
            raise err

        self.add_to_jobs_queue(self.dev.set_wanted_frequency, f_Hz)
        self.add_to_jobs_queue(
            "signal_GUI_input_field_update", GUI_input_fields.F_WANTED
        )
        self.process_jobs_queue()

    @Slot()
    def _process_rbtn_mode(self):
        # Very elaborate logic scheme here, but necessary to have the
        # confirmation dialog work all right.

        # Figure out the request
        if self.rbtn_mode_pressure.isChecked():
            reqs_mode = HVL_Mode.CONTROLLER
            reqs_mode_str = "regulate pressure"
        else:
            reqs_mode = HVL_Mode.ACTUATOR
            reqs_mode_str = "fixed frequency"

        if self.dev.state.hvl_mode == reqs_mode:
            # Actually no change requested
            return

        # Ask user to confirm request
        reply = QtWid.QMessageBox.question(
            self.qgrp_control,
            "Confirm mode change",
            f"Change the mode to `{reqs_mode_str}`?",
            QtWid.QMessageBox.StandardButton.Yes
            | QtWid.QMessageBox.StandardButton.No,
        )

        # Canceled by user
        if reply == QtWid.QMessageBox.StandardButton.No:
            self.rbtn_mode_pressure.setChecked(
                self.dev.state.hvl_mode == HVL_Mode.CONTROLLER
            )
            self.rbtn_mode_frequency.setChecked(
                self.dev.state.hvl_mode == HVL_Mode.ACTUATOR
            )
            return

        # Eager mode change
        self.dev.state.hvl_mode = reqs_mode

        # Send and confirm mode change coming back from the pump
        self.add_to_jobs_queue(self.dev.set_hvl_mode, reqs_mode)
        self.add_to_jobs_queue(
            "signal_GUI_input_field_update", GUI_input_fields.HVL_MODE
        )
        self.process_jobs_queue()

    # --------------------------------------------------------------------------
    #   Worker communication functions
    # --------------------------------------------------------------------------

    @Slot()
    def send_pump_start(self):
        """Schedule a 'pump start' as soon as possible."""
        self.send(self.dev.pump_start)

        # React as fast as possible
        QtWid.QApplication.processEvents()

    @Slot()
    def send_pump_stop(self):
        """Schedule a 'pump stop' as soon as possible."""
        self.send(self.dev.pump_stop)
        self.pump_is_stopping = True

        # React as fast as possible
        QtWid.QApplication.processEvents()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Multithreaded PyQt/PySide GUI to interface with a Keysight N8700 power supply
(PSU).
"""
__author__ = "Dennis van Gils"
__authoremail__ = "vangils.dennis@gmail.com"
__url__ = "https://github.com/Dennis-van-Gils/python-dvg-devices"
__date__ = "12-06-2024"
__version__ = "1.5.0"
print(__url__)
# pylint: disable=wrong-import-position, missing-function-docstring, bare-except

import os
import sys
from pathlib import Path
from typing import List

import qtpy
from qtpy import QtCore, QtGui, QtWidgets as QtWid
import pyvisa

import dvg_pyqt_controls as controls
from dvg_debug_functions import ANSI, dprint
from dvg_qdeviceio import DAQ_TRIGGER
from dvg_devices.Keysight_N8700_protocol_SCPI import Keysight_N8700
from dvg_devices.Keysight_N8700_qdev import Keysight_N8700_qdev

# Show debug info in terminal? Warning: Slow! Do not leave on unintentionally.
DEBUG = False

# ------------------------------------------------------------------------------
#   MainWindow
# ------------------------------------------------------------------------------


class MainWindow(QtWid.QWidget):
    def __init__(
        self,
        qdevs: List[Keysight_N8700_qdev],
        parent=None,
        **kwargs,
    ):
        super().__init__(parent, **kwargs)

        self.setWindowTitle("Keysight N8700 power supply control")
        self.setGeometry(40, 60, 0, 0)
        self.setStyleSheet(
            controls.SS_TEXTBOX_READ_ONLY
            + controls.SS_GROUP
            + controls.SS_HOVER
        )

        # Top grid
        self.lbl_title = QtWid.QLabel("Keysight N8700 power supply control")
        self.lbl_title.setFont(
            QtGui.QFont("Palatino", 14, weight=QtGui.QFont.Weight.Bold)
        )
        self.pbtn_exit = QtWid.QPushButton("Exit")
        self.pbtn_exit.clicked.connect(self.close)
        self.pbtn_exit.setMinimumHeight(30)

        grid_top = QtWid.QGridLayout()
        grid_top.addWidget(self.lbl_title, 0, 0)
        grid_top.addWidget(
            self.pbtn_exit, 0, 1, QtCore.Qt.AlignmentFlag.AlignRight
        )

        # PSU groups
        hbox1 = QtWid.QHBoxLayout()
        for qdev in qdevs:
            hbox1.addWidget(qdev.grpb)
        hbox1.addStretch(1)

        # Round up full window
        vbox = QtWid.QVBoxLayout(self)
        vbox.addLayout(grid_top)
        vbox.addLayout(hbox1)
        vbox.addStretch(1)


# ------------------------------------------------------------------------------
#   Main
# ------------------------------------------------------------------------------

if __name__ == "__main__":
    # VISA addresses of the Keysight PSUs
    VISA_ADDRESS_PSU_1 = "USB0::0x0957::0x8707::US15M3727P::INSTR"
    VISA_ADDRESS_PSU_2 = "USB0::0x0957::0x8707::US15M3728P::INSTR"
    VISA_ADDRESS_PSU_3 = "USB0::0x0957::0x8707::US15M3726P::INSTR"

    # Config files
    PATH_CONFIG_PSU_1 = Path(
        os.getcwd() + "/config/settings_Keysight_PSU_1.txt"
    )
    PATH_CONFIG_PSU_2 = Path(
        os.getcwd() + "/config/settings_Keysight_PSU_2.txt"
    )
    PATH_CONFIG_PSU_3 = Path(
        os.getcwd() + "/config/settings_Keysight_PSU_3.txt"
    )

    # The state of the PSUs is polled with this time interval
    UPDATE_INTERVAL_MS = 1000  # [ms]

    # --------------------------------------------------------------------------
    #   Connect to and set up Keysight power supplies (PSU)
    # --------------------------------------------------------------------------

    rm = pyvisa.ResourceManager()

    psu1 = Keysight_N8700(VISA_ADDRESS_PSU_1, PATH_CONFIG_PSU_1, "PSU 1")
    psu2 = Keysight_N8700(VISA_ADDRESS_PSU_2, PATH_CONFIG_PSU_2, "PSU 2")
    psu3 = Keysight_N8700(VISA_ADDRESS_PSU_3, PATH_CONFIG_PSU_3, "PSU 3")
    psus = [psu1, psu2, psu3]

    for psu in psus:
        if psu.connect(rm):
            psu.read_config_file()
            psu.begin()

    # --------------------------------------------------------------------------
    #   Create application
    # --------------------------------------------------------------------------

    main_thread = QtCore.QThread.currentThread()
    if isinstance(main_thread, QtCore.QThread):
        main_thread.setObjectName("MAIN")  # For DEBUG info

    if qtpy.PYQT6 or qtpy.PYSIDE6:
        sys.argv += ["-platform", "windows:darkmode=0"]
    app = QtWid.QApplication(sys.argv)
    app.setStyle("Fusion")

    # --------------------------------------------------------------------------
    #   Set up communication threads for the PSUs
    # --------------------------------------------------------------------------

    psu_qdevs: List[Keysight_N8700_qdev] = []
    for psu in psus:
        psu_qdevs.append(
            Keysight_N8700_qdev(
                dev=psu,
                DAQ_trigger=DAQ_TRIGGER.SINGLE_SHOT_WAKE_UP,
                debug=DEBUG,
            )
        )

    # DEBUG information
    psu_qdevs[0].worker_DAQ.debug_color = ANSI.YELLOW
    psu_qdevs[0].worker_jobs.debug_color = ANSI.CYAN
    psu_qdevs[1].worker_DAQ.debug_color = ANSI.GREEN
    psu_qdevs[1].worker_jobs.debug_color = ANSI.RED

    for psu_qdev in psu_qdevs:
        psu_qdev.start()

    # --------------------------------------------------------------------------
    #   Set up PSU update timer
    # --------------------------------------------------------------------------

    def trigger_update_psus():
        if DEBUG:
            dprint("timer_psus: wake up all DAQ")

        for psu_qdev_ in psu_qdevs:
            psu_qdev_.wake_up_DAQ()

    timer_psus = QtCore.QTimer()
    timer_psus.timeout.connect(trigger_update_psus)
    timer_psus.start(UPDATE_INTERVAL_MS)

    # --------------------------------------------------------------------------
    #   Start the main GUI event loop
    # --------------------------------------------------------------------------

    def about_to_quit():
        print("About to quit")
        app.processEvents()
        for psu_qdev_ in psu_qdevs:
            psu_qdev_.quit()
        for psu_ in psus:
            try:
                psu_.close()
            except:
                pass
        try:
            rm.close()
        except:
            pass

    app.aboutToQuit.connect(about_to_quit)
    window = MainWindow(qdevs=psu_qdevs)
    window.show()

    sys.exit(app.exec())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""RS232 function library for Thermo Scientific ThermoFlex recirculating
chillers.

Communication errors will be handled as non-fatal. This means it will struggle
on with the script while reporting error messages to the command line output,
as opposed to terminating the program completely.

State variables that read numpy.nan indicate that they are uninitialized or that
the previous query resulted in a communication error.

When this module is directly run from the terminal a demo will be shown.
"""
__author__ = "Dennis van Gils"
__authoremail__ = "vangils.dennis@gmail.com"
__url__ = "https://github.com/Dennis-van-Gils/python-dvg-devices"
__date__ = "23-05-2024"
__version__ = "1.5.0"
# pylint: disable=missing-function-docstring, broad-except

import sys
import time
from typing import Union, Tuple

import numpy as np

from dvg_debug_functions import print_fancy_traceback as pft
from dvg_devices.BaseDevice import SerialDevice

# RS232 header of binary serial communication
RS232_START = b"\xCA\x00\x01"


class Unit_of_measure:
    # fmt: off
    no_unit = 0
    deg_C   = 1     # Temperature in degrees Celsius
    deg_F   = 2     # Temperature in degrees Fahrenheit
    LPM     = 3     # Flow in liters per minute
    GPM     = 4     # Flow in gallons per minute
    sec     = 5     # Time in seconds
    PSI     = 6     # Pressure in pounds per square inch
    bar     = 7     # Pressure in bars  # pylint: disable=disallowed-name
    MOhm_cm = 8     # Resistivity in mega-Ohms per centimeter
    percent = 9     # Percentage
    Volt    = 10    # Voltage in Volts
    kPa     = 11    # Pressure in kilo-Pascals
    # fmt: on


# Units expected to be used by the chiller
CHILLER_FLOW_UNIT = Unit_of_measure.LPM
CHILLER_TEMP_UNIT = Unit_of_measure.deg_C
CHILLER_PRES_UNIT = Unit_of_measure.bar


class ThermoFlex_chiller(SerialDevice):
    """Containers for the process and measurement variables.
    [numpy.nan] values indicate that the parameter is not initialized or that
    the last query was unsuccessful in communication.
    """

    # fmt: off
    class Units:
        # Container for the units used and expected by the chiller
        temp: float = np.nan  # Unit of measure index for the temperature
        flow: float = np.nan  # Unit of measure index for the flow rate
        pres: float = np.nan  # Unit of measure index for the pressure

    class Values_alarm:
        # Container for the alarm values
        LO_temp: float = np.nan  # Low temperature limit  ['C]
        HI_temp: float = np.nan  # High temperature limit ['C]
        LO_flow: float = np.nan  # Low flow rate limit    [LPM]
        HI_flow: float = np.nan  # High flow rate limit   [LPM]
        LO_pres: float = np.nan  # Low pressure limit     [bar]
        HI_pres: float = np.nan  # High pressure limit    [bar]

    class Values_PID:
        # Container for the PID values
        P: float = np.nan        # Proportional term      [% span of 100 'C]
        I: float = np.nan        # Integral term          [repeats/minute]
        D: float = np.nan        # Derivative term        [minutes]

    class Status_bits:
        # Container for the status bits of the chiller
        running                    : Union[int, float] = np.nan
        RTD1_open                  : Union[int, float] = np.nan
        RTD2_open                  : Union[int, float] = np.nan
        RTD3_open                  : Union[int, float] = np.nan
        high_temp_fixed_fault      : Union[int, float] = np.nan
        low_temp_fixed_fault       : Union[int, float] = np.nan
        high_temp_fault            : Union[int, float] = np.nan
        low_temp_fault             : Union[int, float] = np.nan
        high_pressure_fault        : Union[int, float] = np.nan
        low_pressure_fault         : Union[int, float] = np.nan
        phase_monitor_fault        : Union[int, float] = np.nan
        high_level_fault           : Union[int, float] = np.nan
        drip_pan_fault             : Union[int, float] = np.nan
        motor_overload_fault       : Union[int, float] = np.nan
        LPC_fault                  : Union[int, float] = np.nan
        HPC_fault                  : Union[int, float] = np.nan
        external_EMO_fault         : Union[int, float] = np.nan
        local_EMO_fault            : Union[int, float] = np.nan
        low_flow_fault             : Union[int, float] = np.nan
        low_level_fault            : Union[int, float] = np.nan
        sense_5V_fault             : Union[int, float] = np.nan
        invalid_level_fault        : Union[int, float] = np.nan
        low_fixed_flow_warning     : Union[int, float] = np.nan
        high_pressure_fault_factory: Union[int, float] = np.nan
        low_pressure_fault_factory : Union[int, float] = np.nan
        powering_up                : Union[int, float] = np.nan
        powering_down              : Union[int, float] = np.nan
        fault_tripped              : Union[int, float] = np.nan

    class State:
        # Container for the process and measurement variables
        setpoint    : float = np.nan  # Setpoint read out of the chiller         ['C]
        temp        : float = np.nan  # Temperature measured by the chiller      ['C]
        flow        : float = np.nan  # Flow rate measured by the chiller        [LPM]
        supply_pres : float = np.nan  # Supply pressure measured by the chiller  [bar]
        suction_pres: float = np.nan  # Suction pressure measured by the chiller [bar]
    # fmt: on

    # --------------------------------------------------------------------------
    #   __init__
    # --------------------------------------------------------------------------

    def __init__(
        self,
        name: str = "chiller",
        long_name: str = "ThermoFlex chiller",
        min_setpoint_degC: float = 10,
        max_setpoint_degC: float = 40,
    ):
        super().__init__(name=name, long_name=long_name)

        # Default serial settings
        self.serial_settings = {
            "baudrate": 9600,
            "timeout": 1,
            "write_timeout": 1,
        }
        # The chiller does not use any EOL termination characters, hence the
        # device should be given time to have its reply been fully send to the
        # computer's serial-in buffer.
        # 9600 baud is ~ 960 bytes per second
        self.set_read_termination(None, query_wait_time=0.05)
        self.set_write_termination(None)

        self.set_ID_validation_query(
            ID_validation_query=self.ID_validation_query,
            valid_ID_broad=True,
            valid_ID_specific=None,
        )

        # Software limits on the temperature setpoint
        self.min_setpoint_degC = min_setpoint_degC
        self.max_setpoint_degC = max_setpoint_degC

        # Container for the units used and expected by the chiller.
        # Gets updated by calling the alarm value queries (e.g.
        # `query_alarm_LO_flow()` etc.) or by calling `begin()`.
        self.units = self.Units()

        # Container for the alarm values.
        # Gets updated by calling the alarm value queries (e.g.
        # `query_alarm_LO_flow()` etc.) or by calling `begin()`.
        self.values_alarm = self.Values_alarm()

        # Container for the PID values
        # Gets updated by calling the PID queries (e.g. `query_PID_P()` etc.)
        # or by calling `begin()`.
        self.values_PID = self.Values_PID()

        # Container for the status bits (faults and warnings of the chiller).
        # Gets updated by calling `query_status_bits()` or by calling `begin()`.
        self.status_bits = self.Status_bits()

        # Container for the process and measurement variables
        self.state = self.State()

    # --------------------------------------------------------------------------
    #   OVERRIDE: query
    # --------------------------------------------------------------------------

    def query(
        self,
        msg: Union[str, bytes],
        raises_on_timeout: bool = False,
        returns_ascii: bool = False,
    ) -> Tuple[bool, Union[bytes, None]]:
        """Send a message to the ThermoFlex chiller and subsequently read the
        reply.

        Args:
            msg (`str` | `bytes`):
                ASCII string or bytes to be sent to the serial device.

            raises_on_timeout (`bool`, optional):
                Should an exception be raised when a write or read timeout
                occurs?

                Default: `False`

            returns_ascii (`bool`, ignored):
                Argument is ignored and is locked to True.

        Returns:
            success (`bool`):
                True if successful, False otherwise.

            reply (`bytes` | `None`):
                Reply received from the device as bytes. `None` if unsuccessful.
        """
        _success, reply = super().query(
            msg,
            raises_on_timeout,
            returns_ascii=False,  # Binary I/O, not ASCII
        )

        # The ThermoFlex is more complex in its replies than the average device.
        # Hence:
        if isinstance(reply, bytes):
            if (len(reply) >= 4) and reply[3] == 0x0F:
                # Error reported by chiller
                if reply[5] == 1:
                    pft("Bad command received by chiller", 3)
                elif reply[5] == 2:
                    pft("Bad data received by chiller", 3)
                elif reply[5] == 3:
                    pft("Bad checksum received by chiller", 3)
                return False, None

            # We got a reply back from /a/ device, not necessarily a
            # ThermoFlex chiller.
            return True, reply

        return False, None

    # --------------------------------------------------------------------------
    #   ID_validation_query
    # --------------------------------------------------------------------------

    def ID_validation_query(self) -> Tuple[bool, None]:
        return self.query_Ack(), None

    # --------------------------------------------------------------------------
    #   begin
    # --------------------------------------------------------------------------

    def begin(self):
        """This function should run directly after having established a
        connection to a ThermoFlex chiller.
        """
        # Query alarm values and units and check for proper units
        self.query_alarm_values_and_units()

        if self.units.flow != CHILLER_FLOW_UNIT:
            print("WARNING: chiller uses the wrong flowrate unit")
            sys.exit(0)

        if self.units.temp != CHILLER_TEMP_UNIT:
            print("WARNING: chiller uses the wrong temperature unit")
            sys.exit(0)

        if self.units.pres != CHILLER_PRES_UNIT:
            print("WARNING: chiller uses the wrong pressure unit")
            sys.exit(0)

        # Query PID values
        self.query_PID_values()

        # Query status bits
        self.query_status_bits()

        # Query setpoint
        self.query_setpoint()

    # --------------------------------------------------------------------------
    #   Query functions
    # --------------------------------------------------------------------------

    def query_data_as_float_and_uom(
        self, msg_bytes: bytes
    ) -> Tuple[bool, float, Union[int, float]]:
        """Query the ThermoFlex chiller and parse its reply as data bytes
        decoding a float value and an unit of measure index.

        Args:
            msg_bytes (bytes): Message to be sent to the chiller.

        Returns:
            success (`bool`):
                True if successful, False otherwise.

            value (`float`):
                The decoded float value. `numpy.nan` if unsuccessful.

            uom (`int` | `float`):
                Unit of measure index. `numpy.nan` if unsuccessful.
        """
        # print(pretty_bytes_to_hex(msg_bytes))   # Debug info
        _success, reply = self.query(msg_bytes)
        # print(pretty_bytes_to_hex(reply))       # Debug info

        if isinstance(reply, bytes):
            value, uom = self.parse_data_bytes(reply)

            if (not np.isnan(value)) and (not np.isnan(uom)):
                return True, value, uom  # Success

        return False, np.nan, np.nan  # Failed

    # --------------------------------------------------------------------------
    #   Parsing
    # --------------------------------------------------------------------------

    def parse_data_bytes(
        self, ans_bytes: bytes
    ) -> Tuple[float, Union[int, float]]:
        """Parse the data bytes. The manual states:

        data_bytes[0]:
            The qualifier byte, where b.0 to b.3 indicates unit of measure index
            and b.4 to b.7 indicates precision of measurement.

        data_bytes[1:]:
            Value.

        Returns:
            value (`float`):
                The decoded float value. [numpy.nan] if unsuccessful.

            uom (`int` | `float`):
                Unit of measure index. [numpy.nan] if unsuccessful.
        """
        value = np.nan
        uom = np.nan
        try:
            nn = ans_bytes[4]  # Number of data bytes to follow
            data_bytes = ans_bytes[5 : 5 + nn]
            pom = data_bytes[0] >> 4  # Precision of measurement
            uom = data_bytes[0] % 0x10  # Unit of measure index
            int_value = int.from_bytes(
                data_bytes[1:], byteorder="big", signed=False
            )
        except Exception as err:
            pft(err, 3)
        else:
            if pom == 0:
                value = int_value
            elif pom == 1:
                value = int_value * 0.1
            elif pom == 2:
                value = int_value * 0.01
            elif pom == 3:
                value = int_value * 0.001
            elif pom == 4:
                value = int_value * 0.0001

        return value, uom

    def parse_status_bits(self, ans_bytes: bytes):
        """Parse the status bits, which are indicators for any faults and/or
        warnings of the chiller. This status gets stored in the class member
        `status_bits`.
        """
        nn = ans_bytes[4]  # Number of data bytes to follow
        status_bits = ans_bytes[5 : 5 + nn]
        d1 = np.uint8(status_bits[0])
        d2 = np.uint8(status_bits[1])
        d3 = np.uint8(status_bits[2])
        d4 = np.uint8(status_bits[3])

        (
            self.status_bits.low_temp_fault,
            self.status_bits.high_temp_fault,
            self.status_bits.low_temp_fixed_fault,
            self.status_bits.high_temp_fixed_fault,
            self.status_bits.RTD3_open,
            self.status_bits.RTD2_open,
            self.status_bits.RTD1_open,
            self.status_bits.running,
        ) = np.unpackbits(d1)
        (
            self.status_bits.HPC_fault,
            self.status_bits.LPC_fault,
            self.status_bits.motor_overload_fault,
            self.status_bits.phase_monitor_fault,
            self.status_bits.high_level_fault,
            self.status_bits.drip_pan_fault,
            self.status_bits.low_pressure_fault,
            self.status_bits.high_pressure_fault,
        ) = np.unpackbits(d2)
        (
            self.status_bits.high_pressure_fault_factory,
            self.status_bits.low_fixed_flow_warning,
            self.status_bits.invalid_level_fault,
            self.status_bits.sense_5V_fault,
            self.status_bits.low_level_fault,
            self.status_bits.low_flow_fault,
            self.status_bits.local_EMO_fault,
            self.status_bits.external_EMO_fault,
        ) = np.unpackbits(d3)
        (
            _dummy,
            _dummy,
            _dummy,
            _dummy,
            _dummy,
            self.status_bits.powering_down,
            self.status_bits.powering_up,
            self.status_bits.low_pressure_fault_factory,
        ) = np.unpackbits(d4)

        self.status_bits.fault_tripped = int(
            sum(
                [
                    self.status_bits.high_temp_fixed_fault,
                    self.status_bits.low_temp_fixed_fault,
                    self.status_bits.high_temp_fault,
                    self.status_bits.low_temp_fault,
                    d2,
                    d3,
                    self.status_bits.low_pressure_fault_factory,
                ]
            )
            > 0
        )

        # print(np.unpackbits(d1))
        # print(np.unpackbits(d2))
        # print(np.unpackbits(d3))
        # print(np.unpackbits(d4))

    def parse_ASCII_bytes(self, ans_bytes: bytes) -> Union[str, None]:
        """Parse the ASCII-encoded text bytes.

        Returns: The decoded text as `str`, or None if unsuccessful.
        """
        nn = ans_bytes[4]  # Number of data bytes to follow
        ASCII_bytes = ans_bytes[5 : 5 + nn]
        try:
            ans_str = ASCII_bytes.decode("ascii")
        except UnicodeDecodeError:
            return None
        except Exception:
            return None

        return ans_str

    # --------------------------------------------------------------------------
    #   ON/OFF
    # --------------------------------------------------------------------------

    def turn_off(self) -> Union[bool, float]:
        """Turn the chiller off.

        Returns: The effected on/off state of the chiller, or [numpy.nan] if
        unsuccessful.
        """
        msg_bytes = RS232_START + b"\x81\x01\x00\x7C"  # With checksum
        _success, reply = self.query(msg_bytes)
        if isinstance(reply, bytes):
            return bool(reply[5])  # return resulting on/off state

        return np.nan

    def turn_on(self) -> Union[bool, float]:
        """Turn the chiller on.

        Returns: The effected on/off state of the chiller, or [numpy.nan] if
        unsuccessful.
        """
        msg_bytes = RS232_START + b"\x81\x01\x01\x7B"  # With checksum
        _success, reply = self.query(msg_bytes)
        if isinstance(reply, bytes):
            return bool(reply[5])  # return resulting on/off state

        return np.nan

    def query_is_on(self) -> Union[bool, float]:
        """Query the on/off state of the chiller.
        This is identical to the status bit `status.chiller_running`.

        Returns: The on/off state of the chiller, or [numpy.nan] if
        unsuccessful.
        """
        msg_bytes = RS232_START + b"\x81\x01\x02\x7A"  # With checksum
        _success, reply = self.query(msg_bytes)
        if isinstance(reply, bytes):
            return bool(reply[5])  # return resulting on/off state

        return np.nan

    # --------------------------------------------------------------------------
    #   query_Ack
    # --------------------------------------------------------------------------

    def query_Ack(self) -> bool:
        """Query the 'Acknowledge' request (REQ ACK). Basically asking 'Are you
        a ThermoFlex chiller?' to the serial device.

        Returns: True if successful, False otherwise.
        """
        msg_bytes = RS232_START + b"\x00\x00\xFE"  # With checksum
        _success, reply = self.query(msg_bytes)
        if (reply == bytes(RS232_START + b"\x00\x02\x00\x00\xFC")) | (
            reply == bytes(RS232_START + b"\x00\x02\x00\x01\xFB")
        ):
            return True

        return False

    # --------------------------------------------------------------------------
    #   Request HI/LO alarm values
    # --------------------------------------------------------------------------

    def query_alarm_values_and_units(self) -> bool:
        """Query all alarm values and store in the class member `values_alarm`.
        Also stores the units of measure in class member `units`.
        Each will be set to [numpy.nan] if unsuccessful.

        Returns: True if successful, False otherwise.
        """
        return (
            self.query_alarm_LO_flow()
            & self.query_alarm_LO_temp()
            & self.query_alarm_LO_pres()
            & self.query_alarm_HI_flow()
            & self.query_alarm_HI_temp()
            & self.query_alarm_HI_pres()
        )

    def query_alarm_LO_flow(self) -> bool:
        """Query the alarm value and store in the class member `values_alarm`.
        Also stores the unit of measure in class member `units`.
        Both will be set to [numpy.nan] if unsuccessful.

        Returns: True if successful, False otherwise.
        """
        msg_bytes = RS232_START + b"\x30\x00\xCE"  # With checksum
        success, value, units = self.query_data_as_float_and_uom(msg_bytes)
        self.values_alarm.LO_flow = value
        self.units.flow = units
        return success

    def query_alarm_LO_temp(self) -> bool:
        """Query the alarm value and store in the class member `values_alarm`.
        Also stores the unit of measure in class member `units`.
        Will be set to [numpy.nan] if unsuccessful.

        Returns: True if successful, False otherwise.
        """
        msg_bytes = RS232_START + b"\x40\x00\xBE"  # With checksum
        success, value, units = self.query_data_as_float_and_uom(msg_bytes)
        self.values_alarm.LO_temp = value
        self.units.temp = units
        return success

    def query_alarm_LO_pres(self) -> bool:
        """Query the alarm value and store in the class member `values_alarm`.
        Also stores the unit of measure in class member `units`.
        Will be set to [numpy.nan] if unsuccessful.

        Returns: True if successful, False otherwise.
        """
        msg_bytes = RS232_START + b"\x48\x00\xB6"  # With checksum
        success, value, units = self.query_data_as_float_and_uom(msg_bytes)
        self.values_alarm.LO_pres = value
        self.units.pres = units
        return success

    def query_alarm_HI_flow(self) -> bool:
        """Query the alarm value and store in the class member `values_alarm`.
        Also stores the unit of measure in class member `units`.
        Will be set to [numpy.nan] if unsuccessful.

        Returns: True if successful, False otherwise.
        """
        msg_bytes = RS232_START + b"\x50\x00\xAE"  # With checksum
        success, value, units = self.query_data_as_float_and_uom(msg_bytes)
        self.values_alarm.HI_flow = value
        self.units.flow = units
        return success

    def query_alarm_HI_temp(self) -> bool:
        """Query the alarm value and store in the class member `values_alarm`.
        Also stores the unit of measure in class member `units`.
        Will be set to [numpy.nan] if unsuccessful.

        Returns: True if successful, False otherwise.
        """
        msg_bytes = RS232_START + b"\x60\x00\x9E"  # With checksum
        success, value, units = self.query_data_as_float_and_uom(msg_bytes)
        self.values_alarm.HI_temp = value
        self.units.temp = units
        return success

    def query_alarm_HI_pres(self) -> bool:
        """Query the alarm value and store in the class member `values_alarm`.
        Also stores the unit of measure in class member `units`.
        Will be set to [numpy.nan] if unsuccessful.

        Returns: True if successful, False otherwise.
        """
        msg_bytes = RS232_START + b"\x68\x00\x96"  # With checksum
        success, value, units = self.query_data_as_float_and_uom(msg_bytes)
        self.values_alarm.HI_pres = value
        self.units.pres = units
        return success

    # --------------------------------------------------------------------------
    #   Query PID values
    # --------------------------------------------------------------------------

    def query_PID_values(self) -> bool:
        """Query all PID values and store in the class member `values_PID`.
        Each will be set to [numpy.nan] if unsuccessful.

        Returns: True if successful, False otherwise.
        """
        return self.query_PID_P() & self.query_PID_I() & self.query_PID_D()

    def query_PID_P(self) -> bool:
        """Query the PID value and store in the class member `values_PID`.
        Will be set to [numpy.nan] if unsuccessful.

        Returns: True if successful, False otherwise.
        """
        msg_bytes = RS232_START + b"\x74\x00\x8A"  # With checksum
        success, value, _units = self.query_data_as_float_and_uom(msg_bytes)
        self.values_PID.P = value
        return success

    def query_PID_I(self) -> bool:
        """Query the PID value and store in the class member `values_PID`.
        Will be set to [numpy.nan] if unsuccessful.

        Returns: True if successful, False otherwise.
        """
        msg_bytes = RS232_START + b"\x75\x00\x89"  # With checksum
        success, value, _units = self.query_data_as_float_and_uom(msg_bytes)
        self.values_PID.I = value
        return success

    def query_PID_D(self) -> bool:
        """Query the PID value and store in the class member `values_PID`.
        Will be set to [numpy.nan] if unsuccessful.

        Returns: True if successful, False otherwise.
        """
        msg_bytes = RS232_START + b"\x76\x00\x88"  # With checksum
        success, value, _units = self.query_data_as_float_and_uom(msg_bytes)
        self.values_PID.D = value
        return success

    # --------------------------------------------------------------------------
    #   query_status_bits
    # --------------------------------------------------------------------------

    def query_status_bits(self) -> bool:
        """Query and parse the status bits, which are indicators for any faults
        and/or warnings of the chiller. This status gets stored in the class
        member `status_bits`.

        Returns: True if successful, False otherwise.
        """
        msg_bytes = RS232_START + b"\x09\x00\xF5"  # With checksum
        success, reply = self.query(msg_bytes)
        if isinstance(reply, bytes):
            self.parse_status_bits(reply)
        return success

    # --------------------------------------------------------------------------
    #   Query state variables
    # --------------------------------------------------------------------------

    def query_state(self) -> bool:
        """Query all process and measurement variables and store in the class
        member `state`.
        Each will be set to [numpy.nan] if unsuccessful.

        Returns: True if successful, False otherwise.
        """
        return (
            self.query_setpoint()
            & self.query_temp()
            & self.query_flow()
            & self.query_supply_pres()
            & self.query_suction_pres()
        )

    def query_setpoint(self) -> bool:
        """Query and store in the class member `state`:
        The temperature setpoint value read from the chiller.
        Will be set to [numpy.nan] if unsuccessful.

        Returns: True if successful, False otherwise.
        """
        msg_bytes = RS232_START + b"\x70\x00\x8E"  # With checksum
        success, value, _units = self.query_data_as_float_and_uom(msg_bytes)
        self.state.setpoint = value
        return success

    def query_temp(self) -> bool:
        """Query and store in the class member `state`:
        The temperature measured by the chiller.
        Will be set to [numpy.nan] if unsuccessful.

        Returns: True if successful, False otherwise.
        """
        msg_bytes = RS232_START + b"\x20\x00\xDE"  # With checksum
        success, value, _units = self.query_data_as_float_and_uom(msg_bytes)
        self.state.temp = value
        return success

    def query_flow(self) -> bool:
        """Query and store in the class member `state`:
        The flow rate measured by the chiller.
        Will be set to [numpy.nan] if unsuccessful.

        Returns: True if successful, False otherwise.
        """
        msg_bytes = RS232_START + b"\x10\x00\xEE"  # With checksum
        success, value, _units = self.query_data_as_float_and_uom(msg_bytes)
        self.state.flow = value
        return success

    def query_supply_pres(self) -> bool:
        """Query and store in the class member `state`:
        The supply pressure measured by the chiller.
        Will be set to [numpy.nan] if unsuccessful.

        Returns: True if successful, False otherwise.
        """
        msg_bytes = RS232_START + b"\x28\x00\xD6"  # With checksum
        success, value, _units = self.query_data_as_float_and_uom(msg_bytes)
        self.state.supply_pres = value
        return success

    def query_suction_pres(self) -> bool:
        """Query and store in the class member `state`:
        The suction pressure measured by the chiller.
        Will be set to [numpy.nan] if unsuccessful.

        Returns: True if successful, False otherwise.
        """
        msg_bytes = RS232_START + b"\x29\x00\xD5"  # With checksum
        success, value, _units = self.query_data_as_float_and_uom(msg_bytes)
        self.state.suction_pres = value
        return success

    # --------------------------------------------------------------------------
    #   query_display_msg
    # --------------------------------------------------------------------------

    def query_display_msg(self) -> Union[str, None]:
        """Query the display text shown on the chiller.

        Returns: The display text as `str`, or None if unsuccessful.
        """
        msg_bytes = RS232_START + b"\x07\x00\xF7"  # With checksum
        _success, reply = self.query(msg_bytes)
        if isinstance(reply, bytes):
            return self.parse_ASCII_bytes(reply)

        return None

    # --------------------------------------------------------------------------
    #   send_setpoint
    # --------------------------------------------------------------------------

    def send_setpoint(self, temp_deg_C: float) -> bool:
        """Send a new temperature setpoint in [deg C.] to the chiller.
        Subsequently, the chiller replies with the currently set setpoint and
        this value will be stored in the class member `state`.

        Args:
            temp_deg_C (`float`):
                Temperature in [deg C].

        Returns: True if successful, False otherwise.
        """
        if temp_deg_C < self.min_setpoint_degC:
            temp_deg_C = self.min_setpoint_degC
            print(
                "WARNING: setpoint is capped\nto the lower limit of "
                f"{self.min_setpoint_degC:.1f} 'C"
            )
        elif temp_deg_C > self.max_setpoint_degC:
            temp_deg_C = self.max_setpoint_degC
            print(
                "WARNING: setpoint is capped\nto the upper limit of "
                f"{self.max_setpoint_degC:.1f} 'C"
            )

        # Transform temperature to bytes
        pom = 0.1  # Precision of measurement, fixed to 0.1
        temp = int(np.round(temp_deg_C / pom)).to_bytes(2, byteorder="big")
        msg_bytes = add_checksum(RS232_START + b"\xF0\x02" + temp)

        # Send setpoint to chiller and receive the set setpoint
        success, value, _units = self.query_data_as_float_and_uom(msg_bytes)
        self.state.setpoint = value
        return success


# ------------------------------------------------------------------------------
#   add_checksum
# ------------------------------------------------------------------------------


def add_checksum(bytes_in: bytes) -> bytes:
    """Calculate a checksum over the passed `bytes_in` and return the bytes
    extended with the checksum.

    Usage example::

        # Request setpoint
        msg_bytes = RS232_START + b"\x70\x00"
        msg_bytes = add_checksum(msg_bytes)
    """
    # The checksum runs over all bytes, except for the leading byte. It is a
    # bitwise inversion of the 1 byte sum of bytes. We mimic the overflow of the
    # 1 byte sum in Python by using the modulo operator `% 0x100`. The inversion
    # is done by using the XOR operator `^ 0xFF`.

    chksum = (sum(bytes_in[1:]) % 0x100) ^ 0xFF  # Is of tyype `int`
    bytes_out = bytes_in + chksum.to_bytes(length=1, byteorder="big")
    # print(pretty_bytes_to_hex(bytes_out))  # Debug info
    return bytes_out


# ------------------------------------------------------------------------------
#   Debug functions
# ------------------------------------------------------------------------------


def pretty_bytes_to_hex(values: bytes) -> str:
    return " ".join(f"{b:02x}" for b in values)


# -----------------------------------------------------------------------------
#   Main: Will show a demo when run from the terminal
# -----------------------------------------------------------------------------

if __name__ == "__main__":
    import os

    # Path to the config textfile containing the (last used) RS232 port
    PATH_CONFIG = "config/port_ThermoFlex_chiller.txt"

    # Create connection to ThermoFlex chiller over RS232
    chiller = ThermoFlex_chiller()
    if chiller.auto_connect(filepath_last_known_port=PATH_CONFIG):
        chiller.begin()  # Retrieve necessary parameters
    else:
        time.sleep(1)
        sys.exit(0)

    if os.name == "nt":
        import msvcrt

        running_Windows = True
    else:
        running_Windows = False

    # Prepare
    send_setpoint = 22.0
    do_send_setpoint = False

    # Loop
    done = False
    while not done:
        # Check if a new setpoint has to be send
        if do_send_setpoint:
            chiller.send_setpoint(send_setpoint)
            do_send_setpoint = False

        # Measure and report
        chiller.query_status_bits()  # Faults and warnings
        chiller.query_state()  # State variables

        if running_Windows:
            os.system("cls")
            print("Press Q to quit.")
            print("Press S to enter new setpoint.")
            print("Press O to toggle the chiller on/off.")
        else:
            os.system("clear")
            print("Press Control + C to quit.")
            print("No other keyboard input possible because OS is not Windows.")

        print("\n------------------------")
        print("      ALARM VALUES")
        print("        LO  |  HI")
        print(
            f" flow: {chiller.values_alarm.LO_flow:4.1f} | "
            f"{chiller.values_alarm.HI_flow:4.1f}  LPM"
        )
        print(
            f" pres: {chiller.values_alarm.LO_pres:4.2f} | "
            f"{chiller.values_alarm.HI_pres:4.2f}  bar"
        )
        print(
            f" temp: {chiller.values_alarm.LO_temp:4.1f} | "
            f"{chiller.values_alarm.HI_temp:4.1f}  'C"
        )
        print("------------------------")
        print(f" P: {chiller.values_PID.P:4.1f}  %% span 100'C")
        print(f" I: {chiller.values_PID.I:4.2f}  repeats/minute")
        print(f" D: {chiller.values_PID.D:4.1f}  minutes")
        print("------------------------")
        print(f" running         : {chiller.status_bits.running}")
        print(f" powering up/down: {chiller.status_bits.powering_down}")
        print(f" fault_tripped   : {chiller.status_bits.fault_tripped}")
        print(f" MSG: {chiller.query_display_msg()}")
        print("------------------------")
        print(f" setpoint: {chiller.state.setpoint:6.1f} 'C")
        print("------------------------")
        print(f" temp    : {chiller.state.temp:6.1f} 'C")
        print(f" flow    : {chiller.state.flow:6.1f} LPM")
        print(f" supply  : {chiller.state.supply_pres:6.2f} bar")
        print(f" suction : {chiller.state.suction_pres:6.2f} bar")
        print("------------------------")
        sys.stdout.flush()

        # Process keyboard input
        if running_Windows:
            if msvcrt.kbhit():
                key = msvcrt.getch()
                if key == b"q":
                    print("\nAre you sure you want to quit [y/n]?")
                    if msvcrt.getch() == b"y":
                        print("Switching off chiller and quitting.")
                        done = True
                elif key == b"s":
                    input_str = input("\nEnter new setpoint [deg C]: ")
                    try:
                        input_float = float(input_str)
                    except ValueError:
                        print(
                            "Illegal value entered. Setpoint remains unchanged."
                        )
                    else:
                        send_setpoint = input_float
                        do_send_setpoint = True
                elif key == b"o":
                    if chiller.status_bits.running:
                        chiller.turn_off()
                    else:
                        chiller.turn_on()

        # Slow down update period
        time.sleep(0.5)

    chiller.turn_off()
    chiller.close()
    time.sleep(1)
